from __future__ import (
    unicode_literals,
)

from eth_utils import (
    int_to_big_endian,
)
from eth_utils.toolz import (
    compose,
    curry,
)


@curry
def zpad(value, length):
    return value.rjust(length, b"\x00")


zpad32 = zpad(length=32)


int_to_32byte_big_endian = compose(
    zpad32,
    int_to_big_endian,
)



from eth_utils import (
    to_dict,
)

ZERO_32BYTES = b"\x00" * 32
ZERO_ADDRESS = b"\x00" * 20


def yield_key_value_if_value_not_none(key, value):
    if value is not None:
        yield key, value


@to_dict
def make_receipt(
    transaction_hash=ZERO_32BYTES,
    transaction_index=0,
    block_number=0,
    block_hash=ZERO_32BYTES,
    cumulative_gas_used=0,
    blob_gas_used=None,
    blob_gas_price=None,
    _from=ZERO_ADDRESS,
    gas_used=21000,
    effective_gas_price=1000000000,
    contract_address=ZERO_ADDRESS,
    logs=None,
    state_root=b"\x00",
    status=0,
    to=ZERO_ADDRESS,
    _type="0x0",
):
    yield from yield_key_value_if_value_not_none("transaction_hash", transaction_hash)
    yield from yield_key_value_if_value_not_none("transaction_index", transaction_index)
    yield from yield_key_value_if_value_not_none("block_number", block_number)
    yield from yield_key_value_if_value_not_none("block_hash", block_hash)
    yield from yield_key_value_if_value_not_none(
        "cumulative_gas_used", cumulative_gas_used
    )
    yield from yield_key_value_if_value_not_none("gas_used", gas_used)
    yield from yield_key_value_if_value_not_none(
        "effective_gas_price", effective_gas_price
    )
    yield from yield_key_value_if_value_not_none("from", _from)
    yield from yield_key_value_if_value_not_none("to", to)
    yield from yield_key_value_if_value_not_none("type", _type)
    yield from yield_key_value_if_value_not_none("state_root", state_root)
    yield from yield_key_value_if_value_not_none("status", status)
    yield from yield_key_value_if_value_not_none("logs", logs or [])
    yield from yield_key_value_if_value_not_none("contract_address", contract_address)
    yield from yield_key_value_if_value_not_none("blob_gas_used", blob_gas_used)
    yield from yield_key_value_if_value_not_none("blob_gas_price", blob_gas_price)



from __future__ import (
    unicode_literals,
)

from eth.constants import (
    POST_MERGE_DIFFICULTY,
    POST_MERGE_MIX_HASH,
    POST_MERGE_NONCE,
)
from eth.vm.forks import (
    BerlinVM,
    CancunVM,
    FrontierVM,
    GrayGlacierVM,
    LondonVM,
    ParisVM,
    ShanghaiVM,
)
from eth_abi import (
    abi,
)
from eth_account import (
    Account,
)
from eth_typing import (
    HexStr,
)
from eth_utils import (
    ValidationError as EthUtilsValidationError,
    encode_hex,
    is_hexstr,
    to_hex,
    to_wei,
)
import pytest

from eth_tester import (
    EthereumTester,
    PyEVMBackend,
)
from eth_tester.backends.pyevm.main import (
    GENESIS_DIFFICULTY,
    GENESIS_MIX_HASH,
    GENESIS_NONCE,
    generate_genesis_state_for_keys,
    get_default_account_keys,
    get_default_genesis_params,
    setup_tester_chain,
)
from eth_tester.backends.pyevm.utils import (
    is_supported_pyevm_version_available,
)
from eth_tester.exceptions import (
    BlockNotFound,
    ValidationError,
)
from eth_tester.normalization.outbound import (
    normalize_withdrawal,
)
from eth_tester.utils.backend_testing import (
    SIMPLE_TRANSACTION,
    BaseTestBackendDirect,
)

ZERO_ADDRESS_HEX = "0x0000000000000000000000000000000000000000"
MNEMONIC = "test test test test test test test test test test test junk"


@pytest.fixture
def eth_tester():
    if not is_supported_pyevm_version_available():
        pytest.skip("PyEVM is not available")
    backend = PyEVMBackend()
    return EthereumTester(backend=backend)


@pytest.fixture
def accounts_from_mnemonic():
    return [
        "0x1e59ce931B4CFea3fe4B875411e280e173cB7A9C",
        "0xc89D42189f0450C2b2c3c61f58Ec5d628176A1E7",
        "0x318b469BBa396AEc2C60342F9441be36A1945174",
    ]


def test_custom_virtual_machines():
    if not is_supported_pyevm_version_available():
        pytest.skip("PyEVM is not available")

    backend = PyEVMBackend(
        vm_configuration=(
            (0, FrontierVM),
            (3, ParisVM),
        )
    )

    # This should be a FrontierVM block
    VM_at_2 = backend.chain.get_vm_class_for_block_number(2)
    # This should be a ParisVM block
    VM_at_3 = backend.chain.get_vm_class_for_block_number(3)

    assert FrontierVM.__name__ == "FrontierVM"
    assert VM_at_2.__name__ == FrontierVM.__name__

    assert ParisVM.__name__ == "ParisVM"
    assert VM_at_3.__name__ == ParisVM.__name__

    # Right now, just test that EthereumTester doesn't crash
    # Maybe some more sophisticated test to make sure the VMs are set correctly?
    # We should to make sure the VM config translates all the way to the main
    #   tester, maybe with a custom VM that hard-codes some block value? that can
    #   be found with tester.get_block_by_number()?
    EthereumTester(backend=backend)


@pytest.mark.parametrize(
    "vm_class_missing_the_field,vm_class_with_new_field,new_field",
    (
        (BerlinVM, LondonVM, "base_fee_per_gas"),
        (ParisVM, ShanghaiVM, "withdrawals"),
        (ParisVM, ShanghaiVM, "withdrawals_root"),
        (ShanghaiVM, CancunVM, "blob_gas_used"),
        (ShanghaiVM, CancunVM, "excess_blob_gas"),
    ),
)
def test_newly_introduced_block_fields_at_fork_transition(
    vm_class_missing_the_field,
    vm_class_with_new_field,
    new_field,
):
    if not is_supported_pyevm_version_available():
        pytest.skip("PyEVM is not available")

    vm_configuration = ((0, vm_class_missing_the_field), (1, vm_class_with_new_field))
    backend = PyEVMBackend(vm_configuration=vm_configuration)

    # test that the field / key does not exist pre-fork
    with pytest.raises(KeyError):
        backend.get_block_by_number(0)[new_field]

    tester = EthereumTester(backend=backend)

    # Test that outbound block validation doesn't break by getting a block.
    pre_fork_block = tester.get_block_by_number(0)

    # Test that outbound block normalization removes fork-specific field.
    with pytest.raises(KeyError):
        pre_fork_block[new_field]

    # test that the field exists at fork transition
    backend.get_block_by_number("pending")[new_field]
    post_fork_block = tester.get_block_by_number("pending")
    assert post_fork_block[new_field] is not None


def test_london_configuration():
    if not is_supported_pyevm_version_available():
        pytest.skip("PyEVM is not available")

    backend = PyEVMBackend(vm_configuration=((0, LondonVM),))

    assert backend.get_block_by_number(0)["base_fee_per_gas"] == 1000000000

    EthereumTester(backend=backend)


def test_apply_withdrawals():
    if not is_supported_pyevm_version_available():
        pytest.skip("PyEVM is not available")

    backend = PyEVMBackend(vm_configuration=((0, ShanghaiVM),))

    tester = EthereumTester(backend=backend)

    withdrawals = [
        {
            "index": 0,
            "validator_index": 0,
            "address": f"0x{'01' * 20}",
            "amount": 100,
        },
        {
            "index": 2**64 - 1,
            "validator_index": 2**64 - 1,
            "address": b"\x02" * 20,
            "amount": 2**64 - 1,
        },
    ]
    backend.apply_withdrawals(withdrawals)

    mined_block = tester.get_block_by_number("latest")
    assert (
        mined_block["withdrawals"] == normalize_withdrawal(withdrawal)
        for withdrawal in withdrawals
    )
    # withdrawal amounts are in gwei, balance is measured in wei
    assert backend.get_balance(b"\x01" * 20) == 100 * 10**9  # 100 gwei
    assert (
        backend.get_balance(b"\x02" * 20) == (2**64 - 1) * 10**9
    )  # 2**64 - 1 gwei

    assert (
        mined_block["withdrawals_root"]
        == "0xbb49834f60c98815399dfb1a3303cc0f80984c4c7533ecf326bc343d8109127e"
    )


class TestPyEVMBackendDirect(BaseTestBackendDirect):
    def test_generate_custom_genesis_state(self):
        state_overrides = {"balance": to_wei(900000, "ether")}
        invalid_overrides = {"gato": "con botas"}

        # Test creating a specific number of accounts
        account_keys = get_default_account_keys(quantity=2)
        assert len(account_keys) == 2
        account_keys = get_default_account_keys(quantity=10)
        assert len(account_keys) == 10

        # Test the underlying state merging functionality
        genesis_state = generate_genesis_state_for_keys(
            account_keys=account_keys, overrides=state_overrides
        )
        assert len(genesis_state) == len(account_keys) == 10
        for _public_address, account_state in genesis_state.items():
            assert account_state["balance"] == state_overrides["balance"]
            assert account_state["code"] == b""

        # Only existing default genesis state keys can be overridden
        with pytest.raises(ValueError):
            generate_genesis_state_for_keys(
                account_keys=account_keys, overrides=invalid_overrides
            )

        # Use staticmethod state overriding
        genesis_state = PyEVMBackend.generate_genesis_state(
            overrides=state_overrides, num_accounts=3
        )
        assert len(genesis_state) == 3
        for _public_address, account_state in genesis_state.items():
            assert account_state["balance"] == state_overrides["balance"]
            assert account_state["code"] == b""

        # Only existing default genesis state keys can be overridden
        with pytest.raises(ValueError):
            PyEVMBackend.generate_genesis_state(overrides=invalid_overrides)

    def test_override_genesis_state(self):
        state_overrides = {"balance": to_wei(900000, "ether")}
        test_accounts = 3

        # Initialize PyEVM backend with custom genesis state
        genesis_state = PyEVMBackend.generate_genesis_state(
            overrides=state_overrides, num_accounts=test_accounts
        )

        # Test the correct number of accounts are created with the specified
        # balance override
        pyevm_backend = PyEVMBackend(genesis_state=genesis_state)
        assert len(pyevm_backend.account_keys) == test_accounts
        for private_key in pyevm_backend.account_keys:
            account = private_key.public_key.to_canonical_address()
            balance = pyevm_backend.get_balance(account=account)
            assert balance == state_overrides["balance"]

        # Test integration with EthereumTester
        tester = EthereumTester(backend=pyevm_backend)
        for private_key in pyevm_backend.account_keys:
            account = private_key.public_key.to_checksum_address()
            balance = tester.get_balance(account=account)
            assert balance == state_overrides["balance"]

    def test_from_mnemonic(self, accounts_from_mnemonic):
        # Initialize PyEVM backend using MNEMONIC, num_accounts,
        # and state overrides (balance)
        num_accounts = 3
        balance = to_wei(15, "ether")  # Give each account 15 Eth
        pyevm_backend = PyEVMBackend.from_mnemonic(
            MNEMONIC,
            num_accounts=num_accounts,
            genesis_state_overrides={"balance": balance},
        )

        # Test integration with EthereumTester
        tester = EthereumTester(backend=pyevm_backend)

        actual_accounts = tester.get_accounts()
        assert len(actual_accounts) == num_accounts

        assert all(acct in accounts_from_mnemonic for acct in actual_accounts)

        for i in range(0, num_accounts):
            actual = actual_accounts[i]
            expected = accounts_from_mnemonic[i]
            assert actual.lower() == expected.lower()
            assert tester.get_balance(account=actual) == balance

    def test_from_mnemonic_override_hd_path(self, accounts_from_mnemonic):
        # Initialize PyEVM backend using MNEMONIC, num_accounts,
        # and custom hd_path
        num_accounts = 3
        pyevm_backend = PyEVMBackend.from_mnemonic(
            MNEMONIC,
            num_accounts=num_accounts,
            hd_path="m/44'/60'/7'",
        )

        # Each of these accounts stems from the MNEMONIC, but with a different hd_path
        expected_accounts = [
            "0x9aEFA413550e6Ae8690642994310d13dDA248b6b",
            "0xcBA8AFA62949343128FE341C3C7F6b119dF78249",
            "0x2C7DdecbF4555dd2220eF92e21B2912342655845",
        ]

        # Test integration with EthereumTester
        tester = EthereumTester(backend=pyevm_backend)

        actual_accounts = tester.get_accounts()
        assert len(actual_accounts) == num_accounts

        assert not any(acct in accounts_from_mnemonic for acct in actual_accounts)
        assert all(acct in expected_accounts for acct in actual_accounts)

    def test_generate_custom_genesis_parameters(self):
        # Establish parameter overrides, for example a custom genesis gas limit
        param_overrides = {"gas_limit": 4750000}

        # Test the underlying default parameter merging functionality
        genesis_params = get_default_genesis_params(overrides=param_overrides)
        assert genesis_params["gas_limit"] == param_overrides["gas_limit"]

        # Use the staticmethod to generate custom genesis parameters
        genesis_params = PyEVMBackend.generate_genesis_params(param_overrides)
        assert genesis_params["gas_limit"] == param_overrides["gas_limit"]

        # Only existing default genesis parameter keys can be overridden
        invalid_overrides = {"gato": "con botas"}
        with pytest.raises(ValueError):
            PyEVMBackend.generate_genesis_params(overrides=invalid_overrides)

    def test_override_genesis_parameters(self):
        # Establish a custom gas limit
        param_overrides = {
            "gas_limit": 4750000,
        }
        block_one_gas_limit = param_overrides["gas_limit"]

        # Initialize PyEVM backend with custom genesis parameters
        genesis_params = PyEVMBackend.generate_genesis_params(overrides=param_overrides)
        pyevm_backend = PyEVMBackend(genesis_parameters=genesis_params)
        genesis_block = pyevm_backend.get_block_by_number(0)
        assert genesis_block["gas_limit"] == param_overrides["gas_limit"]
        pending_block_one = pyevm_backend.get_block_by_number("pending")
        assert pending_block_one["gas_limit"] == block_one_gas_limit

        # Integrate with EthereumTester
        tester = EthereumTester(backend=pyevm_backend)
        genesis_block = tester.get_block_by_number(0)
        assert genesis_block["gas_limit"] == param_overrides["gas_limit"]
        pending_block_one = tester.get_block_by_number("pending")
        assert pending_block_one["gas_limit"] == block_one_gas_limit

    def test_send_transaction_invalid_from(self, eth_tester):
        accounts = eth_tester.get_accounts()
        assert accounts, "No accounts available for transaction sending"

        with pytest.raises(ValidationError, match=r'No valid "from" key was provided'):
            self._send_and_check_transaction(
                eth_tester, SIMPLE_TRANSACTION, ZERO_ADDRESS_HEX
            )

    def test_pending_block_not_found_when_fetched_by_number(self, eth_tester):
        # assert `latest` block can be fetched by number
        latest_block_num = eth_tester.get_block_by_number("latest")["number"]
        assert isinstance(latest_block_num, int)
        eth_tester.get_block_by_number(latest_block_num)

        # assert `pending` block cannot be fetched by number
        pending_block_num = eth_tester.get_block_by_number("pending")["number"]
        assert isinstance(pending_block_num, int)
        assert pending_block_num == latest_block_num + 1

        with pytest.raises(BlockNotFound):
            eth_tester.get_block_by_number(pending_block_num)

    def test_pyevm_backend_with_custom_vm_configuration_pow_to_pos(self):
        vm_config = (
            (0, GrayGlacierVM),
            (3, ParisVM),
        )

        pyevm_backend = PyEVMBackend(vm_configuration=vm_config)
        tester = EthereumTester(backend=pyevm_backend)

        # assert genesis block was created with pre-merge, PoW genesis values
        genesis_block = tester.get_block_by_number(0)

        assert genesis_block["difficulty"] == GENESIS_DIFFICULTY
        assert genesis_block["nonce"] == encode_hex(GENESIS_NONCE)
        assert genesis_block["mix_hash"] == encode_hex(GENESIS_MIX_HASH)

        tester.mine_blocks(3)

        # assert smooth transition to PoS with expected values
        third_block = tester.get_block_by_number(3)
        assert third_block["difficulty"] == POST_MERGE_DIFFICULTY
        assert third_block["nonce"] == encode_hex(POST_MERGE_NONCE)

        # assert not empty mix_hash
        third_block_mix_hash = third_block["mix_hash"]
        assert is_hexstr(third_block_mix_hash)
        assert third_block_mix_hash != encode_hex(POST_MERGE_MIX_HASH)

    def test_pyevm_backend_with_custom_vm_configuration_post_merge(self):
        vm_config = ((0, ParisVM),)

        _acct_keys, chain = setup_tester_chain(vm_configuration=vm_config)

        # assert genesis block was created with post-merge, PoS genesis values
        genesis_block = chain.get_canonical_block_by_number(0)
        assert genesis_block.header.difficulty == POST_MERGE_DIFFICULTY
        assert genesis_block.header.nonce == POST_MERGE_NONCE
        assert genesis_block.header.mix_hash == POST_MERGE_MIX_HASH

    def test_eth_get_storage_at(self):
        # add storage to accounts in the genesis block
        state_overrides = {
            "storage": {
                1: 1,
                2: 2,
            }
        }

        genesis_state = PyEVMBackend.generate_genesis_state(
            overrides=state_overrides, num_accounts=3
        )
        pyevm_backend = PyEVMBackend(genesis_state=genesis_state)
        tester = EthereumTester(backend=pyevm_backend)

        accounts = tester.get_accounts()
        assert len(accounts) == 3

        for acct in accounts:
            assert tester.get_storage_at(acct, HexStr("0x0")) == f"0x{'00'*32}"
            assert tester.get_storage_at(acct, HexStr("0x1")) == f"0x{'00'*31}01"
            assert tester.get_storage_at(acct, HexStr("0x2")) == f"0x{'00'*31}02"

    # --- cancun network upgrade --- #

    BLOB_TEXT = "We are the music makers, And we are the dreamers of dreams."
    ENCODED_BLOB_TEXT = abi.encode(["string"], [BLOB_TEXT])

    BLOB_TX_FOR_SIGNING = {
        "type": 3,
        "chainId": 1337,
        "value": 0,
        "gas": 21000,
        "maxFeePerGas": 10**10,
        "maxPriorityFeePerGas": 10**10,
        "maxFeePerBlobGas": 10**10,
        "nonce": 0,
    }

    def test_send_raw_transaction_valid_blob_transaction(self, eth_tester):
        pkey = eth_tester.backend.account_keys[0]
        acct = Account.from_key(pkey)

        tx = self.BLOB_TX_FOR_SIGNING.copy()
        tx["from"] = acct.address
        tx["to"] = eth_tester.get_accounts()[1]

        # Blobs contain 4096 32-byte field elements. Subtract the length of the encoded
        # text divided into 32-byte chunks from 4096 and pad the rest with zeros.
        blob_data = (
            b"\x00" * 32 * (4096 - len(self.ENCODED_BLOB_TEXT) // 32)
        ) + self.ENCODED_BLOB_TEXT

        signed = acct.sign_transaction(tx, blobs=[blob_data])
        tx_hash = eth_tester.send_raw_transaction(to_hex(signed.rawTransaction))
        assert eth_tester.get_transaction_by_hash(tx_hash)

    def test_send_raw_transaction_invalid_blob_transaction(self, eth_tester):
        pkey = eth_tester.backend.account_keys[0]
        acct = Account.from_key(pkey)

        tx = self.BLOB_TX_FOR_SIGNING.copy()
        tx["from"] = acct.address
        tx["to"] = eth_tester.get_accounts()[1]

        blob_data = (
            b"\x00" * 32 * (4096 - len(self.ENCODED_BLOB_TEXT) // 32)
        ) + self.ENCODED_BLOB_TEXT[
            :-1
        ]  # only 1 byte short -- invalid

        with pytest.raises(EthUtilsValidationError):
            acct.sign_transaction(tx, blobs=[blob_data])



from __future__ import (
    unicode_literals,
)

UINT256_MIN = 0
UINT256_MAX = 2**256 - 1
UINT64_MAX = 2**64 - 1
UINT8_MAX = 2**8 - 1
UINT2048_MAX = 2**2048 - 1


ZERO_ADDRESS_HEX = "0x0000000000000000000000000000000000000000"
BURN_ADDRESS = "0xdead000000000000000000000000000000000000"


FORK_HOMESTEAD = "FORK_HOMESTEAD"
FORK_DAO = "FORK_DAO"
FORK_SPURIOUS_DRAGON = "FORK_SPURIOUS_DRAGON"
FORK_TANGERINE_WHISTLE = "FORK_TANGERINE_WHISTLE"
FORK_BYZANTIUM = "FORK_BYZANTIUM"
FORK_PETERSBURG = "FORK_PETERSBURG"
FORK_ISTANBUL = "FORK_ISTANBUL"
FORK_MUIR_GLACIER = "FORK_MUIR_GLACIER"
FORK_BERLIN = "FORK_BERLIN"
FORK_LONDON = "FORK_LONDON"
FORK_ARROW_GLACIER = "FORK_ARROW_GLACIER"
FORK_GRAY_GLACIER = "FORK_GRAY_GLACIER"
FORK_PARIS = "FORK_PARIS"

KNOWN_FORKS = {
    FORK_HOMESTEAD,
    FORK_DAO,
    FORK_SPURIOUS_DRAGON,
    FORK_TANGERINE_WHISTLE,
    FORK_BYZANTIUM,
    FORK_PETERSBURG,
    FORK_ISTANBUL,
    FORK_MUIR_GLACIER,
    FORK_BERLIN,
    FORK_LONDON,
    FORK_ARROW_GLACIER,
    FORK_GRAY_GLACIER,
    FORK_PARIS,
}


LATEST = "latest"
EARLIEST = "earliest"
PENDING = "pending"
SAFE = "safe"
FINALIZED = "finalized"
BLOCK_NUMBER_META_VALUES = {
    LATEST,
    EARLIEST,
    PENDING,
    SAFE,
    FINALIZED,
}

#
# SECPK1N
#
SECPK1_P = 2**256 - 2**32 - 977
SECPK1_N = (
    115792089237316195423570985008687907852837564279074904382605163141518161494337
)
SECPK1_A = 0
SECPK1_B = 7
SECPK1_Gx = (
    55066263022277343669578718895168534326250603453777594175500187360389116729240
)
SECPK1_Gy = (
    32670510020758816978083085130507043184471273380659243275938904335757337482424
)
SECPK1_G = (SECPK1_Gx, SECPK1_Gy)


# -- KNOWN TRANSACTION TYPES -- #
LEGACY_TX_TYPE = 0
ACCESS_LIST_TX_TYPE = 1
DYNAMIC_FEE_TX_TYPE = 2
BLOB_TX_TYPE = 3


# -- EIP CONSTANTS -- #

# EIP-1559
DYNAMIC_FEE_TRANSACTION_PARAMS = ("max_fee_per_gas", "max_priority_fee_per_gas")

# EIP-4844
BLOB_TRANSACTION_PARAMS = ("max_fee_per_blob_gas", "blob_versioned_hashes")
GAS_PER_BLOB = 2**17



from eth_utils.toolz import (
    merge,
)


def merge_genesis_overrides(defaults, overrides):
    allowed_fields = set(defaults.keys())
    override_fields = set(overrides.keys())
    unexpected_fields = tuple(sorted(override_fields.difference(allowed_fields)))

    if unexpected_fields:
        err = (
            "The following invalid fields were supplied to "
            f"override default genesis values: {unexpected_fields}."
        )
        raise ValueError(err)

    merged_params = merge(defaults, overrides)
    return merged_params



from __future__ import (
    unicode_literals,
)

import pytest

from eth_tester.utils.filters import (
    Filter,
)


@pytest.fixture
def f():
    _f = Filter(None)
    return _f


def test_filter_starts_empty(f):
    assert not f.get_changes()
    assert not f.get_all()

    # put an item in the filter
    f.add("value-a")

    # verify it has values
    assert f.get_all()
    # get all the changes
    assert f.get_changes()

    # verify that it no longer has changes
    assert not f.get_changes()

    # put another item in the filter
    f.add("value-b")

    # verify that it now has changes again
    assert f.get_changes()

    # verify that it no longer has changes
    assert not f.get_changes()


def test_filter_maintains_ordering(f):
    values = ("value-a", "value-b", "value-c", "value-d", "value-e")
    for item in values[:3]:
        f.add(item)

    assert f.get_changes() == values[:3]
    assert f.get_all() == values[:3]

    for item in values[3:]:
        f.add(item)

    assert f.get_changes() == values[3:]
    assert f.get_all() == values



import os
import sys
import warnings

from eth_tester.utils.module_loading import (
    get_import_path,
    import_string,
)

from .mock import (
    MockBackend,
)
from .pyevm import (
    PyEVMBackend,
    is_supported_pyevm_version_available,
)


def get_chain_backend_class(backend_import_path=None):
    warnings.simplefilter("default")

    if backend_import_path is None:
        if "ETHEREUM_TESTER_CHAIN_BACKEND" in os.environ:
            backend_import_path = os.environ["ETHEREUM_TESTER_CHAIN_BACKEND"]
        elif is_supported_pyevm_version_available():
            backend_import_path = get_import_path(PyEVMBackend)
        else:
            warnings.warn(
                UserWarning(
                    "Ethereum Tester: No backend was explicitly set, and no *full* "
                    "backends were available.  Falling back to the `MockBackend` "
                    "which does not support all EVM functionality.  Please refer to "
                    "the `eth-tester` documentation for information on what "
                    "backends are available and how to set them.  Your py-evm "
                    "package may need to be updated."
                ),
                stacklevel=2,
            )
            backend_import_path = get_import_path(MockBackend)
    return import_string(backend_import_path)


def get_chain_backend(backend_class=None):
    if backend_class is None:
        backend_class = get_chain_backend_class()
    return backend_class()



from __future__ import (
    unicode_literals,
)

from eth_utils import (
    decode_hex,
    encode_hex,
    to_dict,
)
import pytest

from eth_tester.exceptions import (
    ValidationError,
)
from eth_tester.validation import (
    DefaultValidator,
)
from eth_tester.validation.inbound import (
    validate_inbound_withdrawals,
)
from tests.utils import (
    yield_key_value_if_value_not_none,
)


@pytest.fixture
def validator():
    _validator = DefaultValidator()
    return _validator


@pytest.mark.parametrize(
    "timestamp,is_valid",
    (
        (4000001, True),
        (4000010, True),
        ("4000001", False),
        ("4000010", False),
        (4000001.0, False),
        (4000010.0, False),
        (True, False),
        (False, False),
    ),
)
def test_time_travel_input_timestamp_validation(validator, timestamp, is_valid):
    if is_valid:
        validator.validate_inbound_timestamp(timestamp)
    else:
        with pytest.raises(ValidationError):
            validator.validate_inbound_timestamp(timestamp)


@pytest.mark.parametrize(
    "block_number,is_valid",
    (
        (0, True),
        (1, True),
        (-1, False),
        (False, False),
        (True, False),
        ("latest", True),
        ("pending", True),
        ("earliest", True),
        ("safe", True),
        ("finalized", True),
        (2**256, True),
        (b"latest", False),
        (b"pending", False),
        (b"earliest", False),
        (b"safe", False),
        (b"finalized", False),
    ),
)
def test_block_number_input_validation(validator, block_number, is_valid):
    if is_valid:
        validator.validate_inbound_block_number(block_number)
    else:
        with pytest.raises(ValidationError):
            validator.validate_inbound_block_number(block_number)


@pytest.mark.parametrize(
    "block_hash,is_valid",
    (
        (0, False),
        (1, False),
        (-1, False),
        (False, False),
        (True, False),
        (b"", False),
        ("", False),
        ("0" * 32, False),
        ("0x" + "0" * 32, False),
        ("\x00" * 32, False),
        (b"\x00" * 32, False),
        ("0" * 64, True),
        ("0x" + "0" * 64, True),
        (b"0x" + b"0" * 64, False),
    ),
)
def test_block_hash_input_validation(validator, block_hash, is_valid):
    if is_valid:
        validator.validate_inbound_block_hash(block_hash)
    else:
        with pytest.raises(ValidationError):
            validator.validate_inbound_block_hash(block_hash)


def _make_filter_params(from_block=None, to_block=None, address=None, topics=None):
    return {
        "from_block": from_block,
        "to_block": to_block,
        "address": address,
        "topics": topics,
    }


@pytest.mark.parametrize(
    "filter_id,is_valid",
    (
        (-1, False),
        (0, True),
        (1, True),
        ("0x0", False),
        ("0x00", False),
        ("0x1", False),
        ("0x01", False),
        ("0", False),
        ("1", False),
    ),
)
def test_filter_id_input_validation(validator, filter_id, is_valid):
    if is_valid:
        validator.validate_inbound_filter_id(filter_id)
    else:
        with pytest.raises(ValidationError):
            validator.validate_inbound_filter_id(filter_id)


ADDRESS_A = encode_hex(b"\x00" * 19 + b"\x01")
ADDRESS_B = encode_hex(b"\x00" * 19 + b"\x02")
TOPIC_A = encode_hex(b"\x00" * 31 + b"\x01")
TOPIC_B = encode_hex(b"\x00" * 31 + b"\x02")
TOPIC_C = encode_hex(b"\x00" * 30 + b"\x01")
TOPIC_D = encode_hex(b"\x00" * 32 + b"\x01")


@pytest.mark.parametrize(
    "filter_params,is_valid",
    (
        (_make_filter_params(), True),
        (_make_filter_params(from_block=0), True),
        (_make_filter_params(to_block=0), True),
        (_make_filter_params(from_block=-1), False),
        (_make_filter_params(to_block=-1), False),
        (_make_filter_params(from_block=True), False),
        (_make_filter_params(to_block=False), False),
        (_make_filter_params(from_block="0x0"), False),
        (_make_filter_params(to_block="0x0"), False),
        (_make_filter_params(from_block="0x1"), False),
        (_make_filter_params(to_block="0x1"), False),
        (_make_filter_params(address=ADDRESS_A), True),
        (_make_filter_params(address=decode_hex(ADDRESS_A)), False),
        (_make_filter_params(address=[ADDRESS_A, ADDRESS_B]), True),
        (_make_filter_params(address=TOPIC_A), False),
        (_make_filter_params(address=decode_hex(TOPIC_A)), False),
        (_make_filter_params(address=[TOPIC_A, ADDRESS_B]), False),
        (_make_filter_params(topics=[TOPIC_A]), True),
        (_make_filter_params(topics=[TOPIC_A, TOPIC_B]), True),
        (_make_filter_params(topics=[TOPIC_A, None]), True),
        (_make_filter_params(topics=[[TOPIC_A], [TOPIC_B]]), True),
        (_make_filter_params(topics=[TOPIC_A, [TOPIC_B, TOPIC_A]]), True),
        (_make_filter_params(topics=[[TOPIC_A], [TOPIC_B, None]]), True),
        (_make_filter_params(topics=[decode_hex(TOPIC_A)]), True),
        (_make_filter_params(topics=[decode_hex(TOPIC_A), decode_hex(TOPIC_B)]), True),
        (_make_filter_params(topics=[decode_hex(TOPIC_A), None]), True),
        (
            _make_filter_params(topics=[[decode_hex(TOPIC_A)], [decode_hex(TOPIC_B)]]),
            True,
        ),
        (
            _make_filter_params(
                topics=[decode_hex(TOPIC_A), [decode_hex(TOPIC_B), decode_hex(TOPIC_A)]]
            ),
            True,
        ),
        (
            _make_filter_params(
                topics=[[decode_hex(TOPIC_A)], [decode_hex(TOPIC_B), None]]
            ),
            True,
        ),
        (_make_filter_params(topics=[decode_hex(TOPIC_C)]), False),
        (_make_filter_params(topics=[decode_hex(TOPIC_D)]), False),
        (_make_filter_params(topics=[ADDRESS_A]), False),
        (_make_filter_params(topics=[ADDRESS_A, TOPIC_B]), False),
        (_make_filter_params(topics=[[ADDRESS_A], [TOPIC_B]]), False),
    ),
)
def test_filter_params_input_validation(validator, filter_params, is_valid):
    if is_valid:
        validator.validate_inbound_filter_params(**filter_params)
    else:
        with pytest.raises(ValidationError):
            validator.validate_inbound_filter_params(**filter_params)


@to_dict
def _make_transaction(
    blob_versioned_hashes=None,
    chain_id=None,
    _type=None,
    _from=None,
    to=None,
    gas=None,
    gas_price=None,
    max_fee_per_blob_gas=None,
    max_fee_per_gas=None,
    max_priority_fee_per_gas=None,
    value=None,
    data=None,
    nonce=None,
    access_list=None,
    r=None,
    s=None,
    v=None,
):
    yield from yield_key_value_if_value_not_none("type", _type)
    yield from yield_key_value_if_value_not_none("chain_id", chain_id)
    yield from yield_key_value_if_value_not_none("from", _from)
    yield from yield_key_value_if_value_not_none("to", to)
    yield from yield_key_value_if_value_not_none("gas", gas)
    yield from yield_key_value_if_value_not_none("gas_price", gas_price)
    yield from yield_key_value_if_value_not_none("max_fee_per_gas", max_fee_per_gas)
    yield from yield_key_value_if_value_not_none(
        "max_priority_fee_per_gas", max_priority_fee_per_gas
    )
    yield from yield_key_value_if_value_not_none("value", value)
    yield from yield_key_value_if_value_not_none("data", data)
    yield from yield_key_value_if_value_not_none("nonce", nonce)
    yield from yield_key_value_if_value_not_none("access_list", access_list)
    yield from yield_key_value_if_value_not_none("r", r)
    yield from yield_key_value_if_value_not_none("s", s)
    yield from yield_key_value_if_value_not_none("v", v)
    yield from yield_key_value_if_value_not_none(
        "blob_versioned_hashes", blob_versioned_hashes
    )
    yield from yield_key_value_if_value_not_none(
        "max_fee_per_blob_gas", max_fee_per_blob_gas
    )


@pytest.mark.parametrize(
    "txn_internal_type, transaction, is_valid",
    (
        ("send", {}, False),
        ("send", _make_transaction(to=ADDRESS_B, gas=21000), False),
        ("send", _make_transaction(_from=ADDRESS_A, gas=21000), True),
        ("send", _make_transaction(_from=ADDRESS_A, to=ADDRESS_B), False),
        ("send", _make_transaction(_from=ADDRESS_A, to=ADDRESS_B, gas=21000), True),
        ("send", _make_transaction(_from="", to=ADDRESS_B, gas=21000), False),
        ("send", _make_transaction(_from=ADDRESS_A, to="", gas=21000), True),
        ("send", _make_transaction(_from=ADDRESS_A, to=b"", gas=21000), False),
        (
            "send",
            _make_transaction(_from=decode_hex(ADDRESS_A), to=ADDRESS_B, gas=21000),
            False,
        ),
        (
            "send",
            _make_transaction(_from=ADDRESS_A, to=decode_hex(ADDRESS_B), gas=21000),
            False,
        ),
        (
            "send",
            _make_transaction(_from=ADDRESS_A, to="", gas=21000, _type="0x0"),
            True,
        ),
        (
            "send",
            _make_transaction(_from=ADDRESS_A, to="", gas=21000, _type="0x1"),
            True,
        ),
        (
            "send",
            _make_transaction(_from=ADDRESS_A, to="", gas=21000, _type="0x01"),
            True,
        ),
        (
            "send",
            _make_transaction(_from=ADDRESS_A, to="", gas=21000, _type="0x2"),
            True,
        ),
        (
            "send",
            _make_transaction(_from=ADDRESS_A, to="", gas=21000, _type="0x02"),
            True,
        ),
        ("send", _make_transaction(_from=ADDRESS_A, to="", gas=21000, _type=1), True),
        (
            "send",
            _make_transaction(_from=ADDRESS_A, to="", gas=21000, _type="0x4"),
            False,
        ),
        (
            "send",
            _make_transaction(_from=ADDRESS_A, to="", gas=21000, _type="1"),
            False,
        ),
        (
            "send",
            _make_transaction(_from=ADDRESS_A, to="", gas=21000, _type="x1"),
            False,
        ),
        ("send", _make_transaction(_from=ADDRESS_A, to="", gas=21000, value=0), True),
        ("send", _make_transaction(_from=ADDRESS_A, to="", gas=21000, value=-1), False),
        ("send", _make_transaction(_from=ADDRESS_A, to="", gas=21000, data=""), True),
        ("send", _make_transaction(_from=ADDRESS_A, to="", gas=21000, data="0x"), True),
        (
            "send",
            _make_transaction(_from=ADDRESS_A, to="", gas=21000, data="0x0"),
            False,
        ),
        ("send", _make_transaction(_from=ADDRESS_A, to="", gas=21000, nonce=0), True),
        ("send", _make_transaction(_from=ADDRESS_A, to="", gas=21000, nonce=1), True),
        ("send", _make_transaction(_from=ADDRESS_A, to="", gas=21000, nonce=-1), False),
        (
            "send",
            _make_transaction(_from=ADDRESS_A, to="", gas=21000, nonce="0x1"),
            False,
        ),
        (
            "send",
            _make_transaction(_from=ADDRESS_A, to="", gas=21000, nonce="arst"),
            False,
        ),
        (
            "send",
            _make_transaction(_from=ADDRESS_A, to="", gas=21000, nonce=True),
            False,
        ),
        (
            "send",
            _make_transaction(_from=ADDRESS_A, to="", gas=21000, nonce=1.0),
            False,
        ),
        ("send", _make_transaction(_from=ADDRESS_A, to="", gas=21000, nonce=-1), False),
        ("send_signed", _make_transaction(_from=ADDRESS_A, gas=21000), False),
        (
            "send_signed",
            _make_transaction(_from=ADDRESS_A, gas=21000, r=1, s=1, v=1),
            True,
        ),
        (
            "send_signed",
            _make_transaction(_from=ADDRESS_A, gas=21000, r=1, s=1, v=256),
            False,
        ),
        (
            "send_signed",
            _make_transaction(
                _from=ADDRESS_A,
                gas=21000,
                max_fee_per_gas=1000000000,
                max_priority_fee_per_gas=1000000000,
                r=1,
                s=1,
                v=1,
            ),
            True,
        ),
        (  # access list txn
            "send",
            _make_transaction(
                _from=ADDRESS_A,
                to="",
                gas=21000,
                gas_price=10000,
                # properly formatted access list
                access_list=(
                    {
                        "address": "0xde0b295669a9fd93d5f28d9ec85e40f4cb697bae",
                        "storage_keys": (
                            "0x0000000000000000000000000000000000000000000000000000000000000003",  # noqa: E501
                            "0x0000000000000000000000000000000000000000000000000000000000000007",  # noqa: E501
                        ),
                    },
                    {
                        "address": "0xbb9bc244d798123fde783fcc1c72d3bb8c189413",
                        "storage_keys": (),
                    },
                ),
            ),
            True,
        ),
        (
            "send",
            _make_transaction(
                _from=ADDRESS_A,
                to="",
                gas=21000,
                gas_price=10000,
                # improperly formatted access list storage key
                access_list=(
                    {
                        "address": "0xde0b295669a9fd93d5f28d9ec85e40f4cb697bae",
                        "storage_keys": (
                            "3",
                            "0x0000000000000000000000000000000000000000000000000000000000000007",  # noqa: E501
                        ),
                    },
                    {
                        "address": "0xbb9bc244d798123fde783fcc1c72d3bb8c189413",
                        "storage_keys": (),
                    },
                ),
            ),
            False,
        ),
        (
            "send",
            _make_transaction(
                _from=ADDRESS_A,
                to="",
                gas=21000,
                # improperly formatted access list address
                access_list=(
                    {
                        "address": "0xde0b295669a9fd93d5f28d9ec85e40f4cb697bae",
                        "storage_keys": (
                            "0x0000000000000000000000000000000000000000000000000000000000000003",  # noqa: E501
                            "0x0000000000000000000000000000000000000000000000000000000000000007",  # noqa: E501
                        ),
                    },
                    {"address": b"", "storage_keys": ()},
                ),
            ),
            False,
        ),
        (  # dynamic fee txn
            "send",
            _make_transaction(
                _from=ADDRESS_A,
                to="",
                gas=21000,
                max_fee_per_gas=1000000000,
                max_priority_fee_per_gas=1000000000,
                # properly formatted access list
                access_list=(
                    {
                        "address": "0xde0b295669a9fd93d5f28d9ec85e40f4cb697bae",
                        "storage_keys": (
                            "0x0000000000000000000000000000000000000000000000000000000000000003",  # noqa: E501
                            "0x0000000000000000000000000000000000000000000000000000000000000007",  # noqa: E501
                        ),
                    },
                    {
                        "address": "0xbb9bc244d798123fde783fcc1c72d3bb8c189413",
                        "storage_keys": (),
                    },
                ),
            ),
            True,
        ),
        # Cancun
        # `eth_sendTransaction` does not support blob transactions:
        ("send", _make_transaction(blob_versioned_hashes=[]), False),
        ("send", _make_transaction(blob_versioned_hashes=[b""]), False),
        ("send", _make_transaction(blob_versioned_hashes=[b"0x"]), False),
        ("send", _make_transaction(max_fee_per_blob_gas=0), False),
        ("send", _make_transaction(max_fee_per_blob_gas=1), False),
        ("send", _make_transaction(max_fee_per_blob_gas="0x0"), False),
    ),
)
def test_transaction_input_validation(
    validator, txn_internal_type, transaction, is_valid
):
    if is_valid:
        validator.validate_inbound_transaction(transaction, txn_internal_type)
    else:
        with pytest.raises(ValidationError):
            validator.validate_inbound_transaction(transaction, txn_internal_type)


@pytest.mark.parametrize(
    "transaction,is_valid",
    (
        ({}, False),
        (_make_transaction(to=ADDRESS_B), False),
        (_make_transaction(gas=21000), False),
        (_make_transaction(_from=ADDRESS_A), True),
        (_make_transaction(_from=ADDRESS_A, nonce=1), True),
        (_make_transaction(_from=ADDRESS_A, gas=21000), True),
        (_make_transaction(_from=ADDRESS_A, gas=True), False),
        (_make_transaction(_from=ADDRESS_A, to=ADDRESS_B), True),
        (_make_transaction(_from=ADDRESS_A, to=ADDRESS_B, gas=21000), True),
        (_make_transaction(_from=""), False),
        (_make_transaction(_from="", to=ADDRESS_B), False),
        (_make_transaction(_from="", gas=21000), False),
        (_make_transaction(_from="", to=ADDRESS_B, gas=21000), False),
        (_make_transaction(_from=ADDRESS_A, to="", gas=21000), True),
        (_make_transaction(_from=ADDRESS_A, to=""), True),
        (_make_transaction(_from=ADDRESS_A, to=b""), False),
        (
            _make_transaction(_from=decode_hex(ADDRESS_A), to=ADDRESS_B, gas=21000),
            False,
        ),
        (_make_transaction(_from=decode_hex(ADDRESS_A), to=ADDRESS_B), False),
        (
            _make_transaction(_from=ADDRESS_A, to=decode_hex(ADDRESS_B), gas=21000),
            False,
        ),
        (_make_transaction(_from=ADDRESS_A, to=decode_hex(ADDRESS_B)), False),
        (_make_transaction(_from=ADDRESS_A, to="", value=0), True),
        (_make_transaction(_from=ADDRESS_A, to="", value=-1), False),
        (_make_transaction(_from=ADDRESS_A, to="", data=""), True),
        (_make_transaction(_from=ADDRESS_A, to="", data=b""), False),
        (_make_transaction(_from=ADDRESS_A, to="", data="0x"), True),
        (_make_transaction(_from=ADDRESS_A, to="", data=b"0x"), False),
        (_make_transaction(_from=ADDRESS_A, to="", data="0x0"), False),
        (_make_transaction(_from=ADDRESS_A, to="", gas=21000, value=0), True),
        (_make_transaction(_from=ADDRESS_A, to="", gas=21000, value=-1), False),
        (_make_transaction(_from=ADDRESS_A, to="", gas=21000, data=""), True),
        (_make_transaction(_from=ADDRESS_A, to="", gas=21000, data="0x"), True),
        (_make_transaction(_from=ADDRESS_A, to="", gas=21000, data="0x0"), False),
    ),
)
def test_transaction_call_and_estimate_gas_input_validation(
    validator, transaction, is_valid
):
    if is_valid:
        validator.validate_inbound_transaction(transaction, txn_internal_type="call")
        validator.validate_inbound_transaction(
            transaction, txn_internal_type="estimate"
        )
    else:
        with pytest.raises(ValidationError):
            validator.validate_inbound_transaction(
                transaction, txn_internal_type="call"
            )
            validator.validate_inbound_transaction(
                transaction, txn_internal_type="estimate"
            )


@pytest.mark.parametrize(
    "withdrawals,is_valid",
    (
        (
            # valid cases all together
            [
                {
                    "index": 0,
                    "validator_index": 0,
                    "address": b"\x00" * 20,  # bytes address
                    "amount": 0,
                },
                {
                    # limit case for uint64 fields
                    "index": 2**64 - 1,
                    "validator_index": 2**64 - 1,
                    "address": b"\x22" * 20,
                    "amount": 2**64 - 1,
                },
                {
                    "index": 0,
                    "validator_index": 0,
                    "address": f"0x{'22' * 20}",  # hexstr address
                    "amount": 0,
                },
            ],
            True,
        ),
        ({}, False),
        (
            {"index": 0, "validator_index": 0, "address": b"\x00" * 20, "amount": 0},
            False,
        ),
        ([{}], False),
        (
            [  # mixed valid and invalid cases
                {  # valid case
                    "index": 0,
                    "validator_index": 0,
                    "address": b"\x00" * 20,
                    "amount": 0,
                },
                {  # invalid case
                    "index": -1,  # negative index
                    "validator_index": 0,
                    "address": b"\x00" * 20,
                    "amount": 0,
                },
            ],
            False,
        ),
        (
            [
                {
                    "index": 2**64,  # out of range
                    "validator_index": 0,
                    "address": b"\x00" * 20,
                    "amount": 0,
                },
            ],
            False,
        ),
        (
            [
                {
                    "index": 0,
                    "validator_index": 2**64,  # out of range
                    "address": b"\x00" * 20,
                    "amount": 0,
                },
            ],
            False,
        ),
        (
            [
                {
                    "index": 0,
                    "validator_index": 0,
                    "address": b"\x00" * 20,
                    "amount": 2**64,  # out of range
                },
            ],
            False,
        ),
        (
            [
                {
                    "index": 0,
                    "validator_index": 0,
                    "address": b"\x00" * 21,  # not 20 bytes
                    "amount": 0,
                },
            ],
            False,
        ),
        (
            [
                {
                    "index": 0,
                    "validator_index": 0,
                    "address": f"0x{'22' * 19}",  # not 20 bytes
                    "amount": 0,
                },
            ],
            False,
        ),
    ),
)
def test_apply_withdrawals_inbound_dict_validation(withdrawals, is_valid):
    if not is_valid:
        with pytest.raises(ValidationError):
            validate_inbound_withdrawals(withdrawals)

    else:
        validate_inbound_withdrawals(withdrawals)


@pytest.mark.parametrize(
    "value,is_valid",
    (
        ("0x0", True),
        ("0x1", True),
        ("0x22", True),
        ("0x4d2", True),
        (0, False),
        (1, False),
        (-1, False),
        ("1", False),
        ("-0x1", False),
        ("test", False),
        (b"test", False),
    ),
)
def test_validate_inbound_storage_slot(value, is_valid):
    if not is_valid:
        with pytest.raises(
            ValidationError,
            match=(
                "Storage slot must be a hex string representation of a positive "
                f"integer - slot: {value}"
            ),
        ):
            DefaultValidator.validate_inbound_storage_slot(value)
    else:
        DefaultValidator.validate_inbound_storage_slot(value)


@pytest.mark.parametrize(
    "value,is_valid",
    (
        (hex(2**256 - 1), True),
        (hex(2**256), False),
    ),
)
def test_validate_inbound_storage_slot_integer_value_at_limit(value, is_valid):
    if not is_valid:
        with pytest.raises(
            ValidationError,
            match="Value exceeds maximum 256 bit integer size",
        ):
            DefaultValidator.validate_inbound_storage_slot(value)
    else:
        DefaultValidator.validate_inbound_storage_slot(value)



from __future__ import (
    absolute_import,
)

from typing import (
    Any,
    Dict,
    Union,
)

import pkg_resources
from semantic_version import (
    Version,
)


def get_pyevm_version():
    try:
        base_version = pkg_resources.parse_version(
            pkg_resources.get_distribution("py-evm").version
        ).base_version
        return Version(base_version)
    except pkg_resources.DistributionNotFound:
        return None


def is_supported_pyevm_version_available():
    version = get_pyevm_version()
    return version and version >= Version("0.5.0")


# --- network utils --- #


if is_supported_pyevm_version_available():
    from eth.abc import (
        BlockAPI,
    )
else:

    class BlockAPI:
        pass


def is_london_block(block: Union[Dict[str, Any], BlockAPI]) -> bool:
    if isinstance(block, BlockAPI):
        try:
            # it's not enough to check hasattr because the attribute could be
            # defined in earlier VM's while raising an AttributeError until implemented
            return block.header.base_fee_per_gas is not None
        except AttributeError:
            return False

    elif isinstance(block, dict) and "base_fee_per_gas" in block:
        return True

    return False


def is_shanghai_block(block: Union[Dict[str, Any], BlockAPI]) -> bool:
    if isinstance(block, BlockAPI):
        try:
            # it's not enough to check hasattr because the attribute could be
            # defined in earlier VM's while raising an AttributeError until implemented
            return block.header.withdrawals_root is not None
        except AttributeError:
            return False
    elif isinstance(block, dict) and "withdrawals_root" in block:
        return True

    return False


def is_cancun_block(block: Union[Dict[str, Any], BlockAPI]) -> bool:
    if isinstance(block, BlockAPI):
        try:
            # it's not enough to check hasattr because the attribute could be
            # defined in earlier VM's while raising an AttributeError until implemented
            return block.header.parent_beacon_block_root is not None
        except AttributeError:
            return False

    elif isinstance(block, dict) and all(
        cancun_field in block
        for cancun_field in (
            "parent_beacon_block_root",
            "blob_gas_used",
            "excess_blob_gas",
        )
    ):
        return True

    return False



from __future__ import (
    absolute_import,
)

from eth_utils import (
    is_integer,
)
from eth_utils.curried import (
    apply_formatter_if,
    apply_one_of_formatters,
    encode_hex,
    is_address,
    is_bytes,
    is_canonical_address,
    is_dict,
    to_checksum_address,
)
from eth_utils.toolz import (
    compose,
    identity,
    partial,
)
from toolz import (
    assoc,
    dissoc,
)

from ..utils.encoding import (
    int_to_32byte_big_endian,
)
from .common import (
    normalize_array,
    normalize_dict,
    normalize_if,
)

normalize_account = to_checksum_address
normalize_account_list = partial(normalize_array, normalizer=normalize_account)

to_empty_or_checksum_address = apply_one_of_formatters(
    (
        (lambda addr: addr == b"", lambda addr: ""),
        (is_canonical_address, to_checksum_address),
    )
)
to_hex_if_integer = apply_formatter_if(is_integer, hex)


def _normalize_outbound_access_list(access_list):
    return tuple(
        [
            {
                "address": to_checksum_address(entry[0]),
                "storage_keys": tuple(
                    [encode_hex(int_to_32byte_big_endian(k)) for k in entry[1]]
                ),
            }
            for entry in access_list
        ]
    )


TRANSACTION_NORMALIZERS = {
    "type": to_hex_if_integer,
    "blob_versioned_hashes": partial(
        normalize_array,
        normalizer=partial(
            normalize_if, conditional_fn=is_bytes, normalizer=encode_hex
        ),
    ),
    "chain_id": identity,
    "hash": encode_hex,
    "nonce": identity,
    "block_hash": partial(normalize_if, conditional_fn=is_bytes, normalizer=encode_hex),
    "block_number": identity,
    "transaction_index": identity,
    "from": to_checksum_address,
    "to": to_empty_or_checksum_address,
    "value": identity,
    "gas": identity,
    "gas_price": identity,
    "max_fee_per_blob_gas": identity,
    "max_fee_per_gas": identity,
    "max_priority_fee_per_gas": identity,
    "data": encode_hex,
    "access_list": _normalize_outbound_access_list,
    "r": identity,
    "s": identity,
    "v": identity,
    "y_parity": identity,
}
normalize_transaction = partial(normalize_dict, normalizers=TRANSACTION_NORMALIZERS)


WITHDRAWAL_NORMALIZERS = {
    "index": identity,
    "validator_index": identity,
    "address": to_checksum_address,
    "amount": identity,
}
normalize_withdrawal = partial(normalize_dict, normalizers=WITHDRAWAL_NORMALIZERS)


def is_transaction_hash_list(value):
    return all(is_bytes(item) for item in value)


def is_transaction_object_list(value):
    return all(is_dict(item) for item in value)


def _remove_fork_specific_fields_if_none(block):
    """
    A `None` value is set for keys if they are not present during outbound block
    validation. This means we are in a VM that has not yet been exposed to this new
    field. Pop this value out here to normalize these older VM blocks.
    """
    for key, value in list(block.items()):
        if value is None:
            block = dissoc(block, key)
    return block


BLOCK_NORMALIZERS = {
    "number": identity,
    "hash": encode_hex,
    "parent_hash": encode_hex,
    "nonce": encode_hex,
    "base_fee_per_gas": identity,
    "sha3_uncles": encode_hex,
    "logs_bloom": identity,
    "transactions_root": encode_hex,
    "receipts_root": encode_hex,
    "state_root": encode_hex,
    "coinbase": to_checksum_address,
    "difficulty": identity,
    "mix_hash": encode_hex,
    "total_difficulty": identity,
    "size": identity,
    "extra_data": encode_hex,
    "gas_limit": identity,
    "gas_used": identity,
    "timestamp": identity,
    "transactions": compose(
        partial(
            normalize_if,
            conditional_fn=is_transaction_hash_list,
            normalizer=partial(normalize_array, normalizer=encode_hex),
        ),
        partial(
            normalize_if,
            conditional_fn=is_transaction_object_list,
            normalizer=partial(normalize_array, normalizer=normalize_transaction),
        ),
    ),
    "uncles": partial(normalize_array, normalizer=encode_hex),
    "withdrawals": partial(normalize_array, normalizer=normalize_withdrawal),
    "withdrawals_root": encode_hex,
    "parent_beacon_block_root": encode_hex,
    "blob_gas_used": identity,
    "excess_blob_gas": identity,
}
normalize_block = compose(
    partial(normalize_dict, normalizers=BLOCK_NORMALIZERS),
    _remove_fork_specific_fields_if_none,
)


LOG_ENTRY_NORMALIZERS = {
    "type": identity,
    "log_index": identity,
    "transaction_index": partial(
        normalize_if,
        conditional_fn=is_bytes,
        normalizer=encode_hex,
    ),
    "transaction_hash": encode_hex,
    "block_hash": partial(
        normalize_if,
        conditional_fn=is_bytes,
        normalizer=encode_hex,
    ),
    "block_number": identity,
    "address": to_checksum_address,
    "data": encode_hex,
    "topics": partial(normalize_array, normalizer=encode_hex),
}
normalize_log_entry = partial(normalize_dict, normalizers=LOG_ENTRY_NORMALIZERS)


def _normalize_contract_address(receipt):
    if receipt["status"] == 0:
        return assoc(receipt, "contract_address", None)
    elif is_address(receipt["contract_address"]):
        return assoc(
            receipt,
            "contract_address",
            to_checksum_address(receipt["contract_address"]),
        )
    else:
        return receipt


RECEIPT_NORMALIZERS = {
    "transaction_hash": encode_hex,
    "transaction_index": identity,
    "block_number": identity,
    "block_hash": partial(
        normalize_if,
        conditional_fn=is_bytes,
        normalizer=encode_hex,
    ),
    "cumulative_gas_used": identity,
    "effective_gas_price": identity,
    "from": to_checksum_address,
    "gas_used": identity,
    "contract_address": identity,  # special case, see ``_normalize_contract_address()``
    "logs": partial(normalize_array, normalizer=normalize_log_entry),
    "state_root": identity,
    "status": identity,
    "to": to_empty_or_checksum_address,
    "type": to_hex_if_integer,
    "base_fee_per_gas": identity,
    "blob_gas_used": identity,
    "blob_gas_price": identity,
}
normalize_receipt = compose(
    partial(normalize_dict, normalizers=RECEIPT_NORMALIZERS),
    _normalize_contract_address,
)



from __future__ import (
    absolute_import,
)

import os

from eth_tester.utils.module_loading import (
    get_import_path,
    import_string,
)

from .default import (
    DefaultNormalizer,
)

DEFAULT_NORMALIZER_CLASS = get_import_path(DefaultNormalizer)


def get_normalizer_backend_class(backend_import_path=None):
    if backend_import_path is None:
        backend_import_path = os.environ.get(
            "ETHEREUM_TESTER_NORMALIZER_BACKEND",
            DEFAULT_NORMALIZER_CLASS,
        )
    return import_string(backend_import_path)


def get_normalizer_backend(backend_class=None):
    if backend_class is None:
        backend_class = get_normalizer_backend_class()
    return backend_class()



from __future__ import (
    unicode_literals,
)

from eth_utils import (
    encode_hex,
)
import pytest
from toolz import (
    dissoc,
    merge,
)

from eth_tester.exceptions import (
    ValidationError,
)
from eth_tester.validation import (
    DefaultValidator,
)
from tests.utils import (
    make_receipt,
)


@pytest.fixture
def validator():
    _validator = DefaultValidator()
    return _validator


@pytest.mark.parametrize(
    "block_hash,is_valid",
    (
        (1, False),
        (True, False),
        (b"\x00" * 32, True),
        (b"\xff" * 32, True),
        ("\x00" * 32, False),
        (encode_hex(b"\x00" * 32), False),
    ),
)
def test_block_hash_output_validation(validator, block_hash, is_valid):
    if is_valid:
        validator.validate_outbound_block_hash(block_hash)
    else:
        with pytest.raises(ValidationError):
            validator.validate_outbound_block_hash(block_hash)


ZERO_32BYTES = b"\x00" * 32
ZERO_8BYTES = b"\x00" * 8
ZERO_ADDRESS = b"\x00" * 20


ADDRESS_A = b"\x00" * 19 + b"\x01"
TOPIC_A = b"\x00" * 31 + b"\x01"
TOPIC_B = b"\x00" * 31 + b"\x02"
HASH32_AS_TEXT = "\x00" * 32
HASH31 = b"\x00" * 31


def _make_legacy_txn(
    hash=ZERO_32BYTES,
    nonce=0,
    block_hash=ZERO_32BYTES,
    block_number=0,
    transaction_index=0,
    _from=ZERO_ADDRESS,
    to=ZERO_ADDRESS,
    value=0,
    gas=21000,
    gas_price=1,
    data=b"",
    v=0,
    r=0,
    s=0,
):
    return {
        "type": "0x0",
        "hash": hash,
        "nonce": nonce,
        "block_hash": block_hash,
        "block_number": block_number,
        "transaction_index": transaction_index,
        "from": _from,
        "to": to,
        "value": value,
        "gas": gas,
        "gas_price": gas_price,
        "data": data,
        "s": s,
        "r": r,
        "v": v,
    }


def _make_access_list_txn(
    chain_id=131277322940537,
    access_list=(),
    **kwargs,
):
    legacy_kwargs = dissoc(dict(**kwargs), "chain_id", "access_list")
    return merge(
        _make_legacy_txn(**legacy_kwargs),
        {
            "type": "0x1",
            "chain_id": chain_id,
            "access_list": list(access_list),
            "y_parity": legacy_kwargs.get("v", 0),
        },
    )


# This is an outbound transaction so we still keep the gas_price for now since the
# gas_price is the min(max_fee_per_gas, base_fee_per_gas + max_priority_fee_per_gas).
# TODO: Sometime in 2022 the inclusion of gas_price may be removed from dynamic fee
#  transactions and we can get rid of this behavior.
#  https://github.com/ethereum/execution-specs/pull/251
def _make_dynamic_fee_txn(
    chain_id=131277322940537,
    max_fee_per_gas=2000000000,
    max_priority_fee_per_gas=1000000000,
    access_list=(),
    **kwargs,
):
    legacy_kwargs = dissoc(
        dict(**kwargs),
        "chain_id",
        "max_fee_per_gas",
        "max_priority_fee_per_gas",
        "access_list",
    )
    return merge(
        _make_access_list_txn(
            chain_id=chain_id, access_list=access_list, **legacy_kwargs
        ),
        {
            "type": "0x2",
            "max_fee_per_gas": max_fee_per_gas,
            "max_priority_fee_per_gas": max_priority_fee_per_gas,
        },
    )


@pytest.mark.parametrize(
    "transaction,is_valid",
    (
        (_make_legacy_txn(), True),
        (_make_access_list_txn(), True),
        (_make_dynamic_fee_txn(), True),
        (_make_legacy_txn(hash=HASH32_AS_TEXT), False),
        (_make_legacy_txn(hash=HASH31), False),
        (_make_legacy_txn(nonce=-1), False),
        (_make_legacy_txn(nonce=1.0), False),
        (_make_legacy_txn(nonce=True), False),
        (_make_legacy_txn(value=-1), False),
        (_make_legacy_txn(value=1.0), False),
        (_make_legacy_txn(value=True), False),
        (_make_legacy_txn(block_number=-1), False),
        (_make_legacy_txn(block_number=1.0), False),
        (_make_legacy_txn(block_number=True), False),
        (_make_legacy_txn(gas=-1), False),
        (_make_legacy_txn(gas=1.0), False),
        (_make_legacy_txn(gas=True), False),
        (_make_legacy_txn(gas_price=-1), False),
        (_make_legacy_txn(gas_price=1.0), False),
        (_make_legacy_txn(gas_price=True), False),
        (_make_legacy_txn(data=""), False),
        (_make_legacy_txn(data="0x"), False),
        (_make_legacy_txn(block_hash=HASH32_AS_TEXT), False),
        (_make_legacy_txn(block_hash=HASH31), False),
        (
            _make_legacy_txn(
                transaction_index=None, block_hash=None, block_number=None
            ),
            True,
        ),
        (
            _make_access_list_txn(
                transaction_index=None,
                block_hash=None,
                block_number=None,
            ),
            True,
        ),
        (
            _make_dynamic_fee_txn(
                transaction_index=None,
                block_hash=None,
                block_number=None,
            ),
            True,
        ),
        (_make_access_list_txn(chain_id="1"), False),
        (_make_dynamic_fee_txn(chain_id="1"), False),
        (_make_access_list_txn(chain_id=-1), False),
        (_make_dynamic_fee_txn(chain_id=-1), False),
        (_make_access_list_txn(chain_id=None), False),
        (_make_dynamic_fee_txn(chain_id=None), False),
        (_make_legacy_txn(v=0), True),
        (_make_dynamic_fee_txn(v=0), True),
        (_make_access_list_txn(v=0), True),
        (_make_legacy_txn(v=1), True),
        (_make_dynamic_fee_txn(v=1), True),
        (_make_access_list_txn(v=1), True),
        (_make_legacy_txn(v=27), True),
        (_make_access_list_txn(v=27), False),
        (_make_dynamic_fee_txn(v=27), False),
        (_make_dynamic_fee_txn(max_fee_per_gas=1.0), False),
        (_make_dynamic_fee_txn(max_priority_fee_per_gas=1.0), False),
        (_make_dynamic_fee_txn(max_fee_per_gas="1"), False),
        (_make_dynamic_fee_txn(max_priority_fee_per_gas="1"), False),
        (
            _make_access_list_txn(
                access_list=((b"\xf0" * 20, (0, 2)),),
            ),
            True,
        ),
        (
            _make_dynamic_fee_txn(
                access_list=((b"\xef" * 20, (1, 2, 3, 4)),),
            ),
            True,
        ),
        (_make_access_list_txn(access_list=()), True),
        (_make_dynamic_fee_txn(access_list=()), True),
        (
            _make_access_list_txn(
                access_list=((b"\xf0" * 19, (0, 2)),),
            ),
            False,
        ),
        (
            _make_dynamic_fee_txn(
                access_list=((b"\xf0" * 19, ()),),
            ),
            False,
        ),
        (
            _make_access_list_txn(
                access_list=((b"\xf0" * 20, ("0", 2)),),
            ),
            False,
        ),
        (
            _make_dynamic_fee_txn(
                access_list=((b"\xf0" * 20, (b"", 1)),),
            ),
            False,
        ),
        (
            _make_dynamic_fee_txn(
                access_list=(("", (1, 2)),),
            ),
            False,
        ),
    ),
)
def test_transaction_output_validation(validator, transaction, is_valid):
    if is_valid:
        validator.validate_outbound_transaction(transaction)
    else:
        with pytest.raises(ValidationError):
            validator.validate_outbound_transaction(transaction)


def _make_log(
    _type="mined",
    log_index=0,
    transaction_index=0,
    transaction_hash=ZERO_32BYTES,
    block_hash=ZERO_32BYTES,
    block_number=0,
    address=ZERO_ADDRESS,
    data=b"",
    topics=None,
):
    return {
        "type": _type,
        "log_index": log_index,
        "transaction_index": transaction_index,
        "transaction_hash": transaction_hash,
        "block_hash": block_hash,
        "block_number": block_number,
        "address": address,
        "data": data,
        "topics": topics or [],
    }


@pytest.mark.parametrize(
    "log_entry,is_valid",
    (
        (_make_log(), True),
        (
            _make_log(
                _type="pending",
                transaction_index=None,
                block_hash=None,
                block_number=None,
            ),
            True,
        ),
        (_make_log(_type="invalid-type"), False),
        (_make_log(transaction_index=-1), False),
        (_make_log(block_number=-1), False),
        (_make_log(transaction_hash=HASH31), False),
        (_make_log(transaction_hash=HASH32_AS_TEXT), False),
        (_make_log(block_hash=HASH31), False),
        (_make_log(block_hash=HASH32_AS_TEXT), False),
        (_make_log(address=encode_hex(ADDRESS_A)), False),
        (_make_log(data=""), False),
        (_make_log(data=None), False),
        (_make_log(topics=[HASH32_AS_TEXT]), False),
        (_make_log(topics=[HASH31]), False),
        (_make_log(topics=[TOPIC_A, TOPIC_B]), True),
        (_make_log(address=ADDRESS_A), True),
    ),
)
def test_log_entry_output_validation(validator, log_entry, is_valid):
    if is_valid:
        validator.validate_outbound_log_entry(log_entry)
    else:
        with pytest.raises(ValidationError):
            validator.validate_outbound_log_entry(log_entry)


@pytest.mark.parametrize(
    "receipt,is_valid",
    (
        (make_receipt(), True),
        (make_receipt(transaction_hash=HASH32_AS_TEXT), False),
        (make_receipt(transaction_hash=HASH31), False),
        (make_receipt(block_hash=HASH32_AS_TEXT), False),
        (make_receipt(block_hash=HASH31), False),
        (make_receipt(transaction_index=-1), False),
        (make_receipt(transaction_index=1.0), False),
        (make_receipt(transaction_index=True), False),
        (make_receipt(block_number=-1), False),
        (make_receipt(block_number=1.0), False),
        (make_receipt(block_number=True), False),
        (make_receipt(gas_used=-1), False),
        (make_receipt(gas_used=1.0), False),
        (make_receipt(gas_used=True), False),
        (make_receipt(cumulative_gas_used=-1), False),
        (make_receipt(cumulative_gas_used=1.0), False),
        (make_receipt(cumulative_gas_used=True), False),
        (make_receipt(contract_address=ZERO_ADDRESS), True),
        (make_receipt(contract_address=encode_hex(ZERO_ADDRESS)), False),
        (make_receipt(logs=[_make_log()]), True),
        (make_receipt(logs=[_make_log(_type="invalid")]), False),
        (make_receipt(_from=ADDRESS_A, to=ADDRESS_A), True),
        (make_receipt(_from=ADDRESS_A, to=ZERO_ADDRESS), True),
        (make_receipt(_from=encode_hex(ZERO_ADDRESS), to=ADDRESS_A), False),
        (make_receipt(_from=ADDRESS_A, to=encode_hex(ZERO_ADDRESS)), False),
        (make_receipt(status=0), True),
        (make_receipt(status=1), True),
        (make_receipt(status=2), False),
        (make_receipt(status=-1), False),
        (make_receipt(blob_gas_used=-1, blob_gas_price=-1), False),
        (make_receipt(blob_gas_used=-1, blob_gas_price=0), False),
        (make_receipt(blob_gas_used=0, blob_gas_price=-1), False),
        (make_receipt(blob_gas_used=0, blob_gas_price=0), True),
        (make_receipt(blob_gas_used=0, blob_gas_price=1), True),
        (make_receipt(blob_gas_used=1, blob_gas_price=0), True),
        (make_receipt(blob_gas_used=1, blob_gas_price=1), True),
        (make_receipt(blob_gas_used=2, blob_gas_price=1), True),
        (make_receipt(blob_gas_used=1, blob_gas_price=2), True),
        (make_receipt(blob_gas_used=2, blob_gas_price=2), True),
    ),
)
def test_receipt_output_validation(validator, receipt, is_valid):
    if is_valid:
        validator.validate_outbound_receipt(receipt)
    else:
        with pytest.raises(ValidationError):
            validator.validate_outbound_receipt(receipt)


def _make_block(
    number=0,
    hash=ZERO_32BYTES,
    parent_hash=ZERO_32BYTES,
    nonce=ZERO_8BYTES,
    sha3_uncles=ZERO_32BYTES,
    logs_bloom=0,
    transactions_root=ZERO_32BYTES,
    receipts_root=ZERO_32BYTES,
    state_root=ZERO_32BYTES,
    coinbase=ZERO_ADDRESS,
    difficulty=0,
    mix_hash=ZERO_32BYTES,
    total_difficulty=0,
    size=0,
    extra_data=ZERO_32BYTES,
    gas_limit=30029122,  # gas limit at London fork block 12965000 on mainnet
    gas_used=21000,
    timestamp=4000000,
    transactions=None,
    uncles=None,
    base_fee_per_gas=1000000000,
    withdrawals=None,
    withdrawals_root=ZERO_32BYTES,
):
    block = {
        "number": number,
        "hash": hash,
        "parent_hash": parent_hash,
        "nonce": nonce,
        "sha3_uncles": sha3_uncles,
        "logs_bloom": logs_bloom,
        "transactions_root": transactions_root,
        "receipts_root": receipts_root,
        "state_root": state_root,
        "coinbase": coinbase,
        "difficulty": difficulty,
        "mix_hash": mix_hash,
        "total_difficulty": total_difficulty,
        "size": size,
        "extra_data": extra_data,
        "gas_limit": gas_limit,
        "gas_used": gas_used,
        "timestamp": timestamp,
        "transactions": transactions or [],
        "uncles": uncles or [],
        "base_fee_per_gas": base_fee_per_gas,
        "withdrawals": withdrawals or [],
        "withdrawals_root": withdrawals_root,
    }
    return block


def _make_withdrawal(
    index: int = 2**64 - 1,
    validator_index: int = 2**64 - 1,
    address: bytes = ZERO_ADDRESS,
    amount: int = 2**64 - 1,
):
    return {
        "index": index,
        "validator_index": validator_index,
        "address": address,
        "amount": amount,
    }


@pytest.mark.parametrize(
    "block,is_valid",
    (
        (_make_block(), True),
        (_make_block(number=-1), False),
        (_make_block(number=1.0), False),
        (_make_block(number=True), False),
        (_make_block(hash=HASH32_AS_TEXT), False),
        (_make_block(hash=HASH31), False),
        (_make_block(parent_hash=HASH32_AS_TEXT), False),
        (_make_block(parent_hash=HASH31), False),
        (_make_block(nonce=-1), False),
        (_make_block(nonce=1.0), False),
        (_make_block(nonce=True), False),
        (_make_block(sha3_uncles=HASH32_AS_TEXT), False),
        (_make_block(logs_bloom=-1), False),
        (_make_block(logs_bloom=1.0), False),
        (_make_block(logs_bloom=True), False),
        (_make_block(transactions_root=HASH32_AS_TEXT), False),
        (_make_block(transactions_root=HASH31), False),
        (_make_block(receipts_root=HASH32_AS_TEXT), False),
        (_make_block(receipts_root=HASH31), False),
        (_make_block(state_root=HASH32_AS_TEXT), False),
        (_make_block(state_root=HASH31), False),
        (_make_block(coinbase=encode_hex(ADDRESS_A)), False),
        (_make_block(difficulty=-1), False),
        (_make_block(difficulty=1.0), False),
        (_make_block(difficulty=True), False),
        (_make_block(mix_hash=HASH32_AS_TEXT), False),
        (_make_block(mix_hash=HASH31), False),
        (_make_block(total_difficulty=-1), False),
        (_make_block(total_difficulty=1.0), False),
        (_make_block(total_difficulty=True), False),
        (_make_block(size=-1), False),
        (_make_block(size=1.0), False),
        (_make_block(size=True), False),
        (_make_block(extra_data=HASH32_AS_TEXT), False),
        (_make_block(extra_data=HASH31), False),
        (_make_block(gas_limit=-1), False),
        (_make_block(gas_limit=1.0), False),
        (_make_block(gas_limit=True), False),
        (_make_block(gas_used=-1), False),
        (_make_block(gas_used=1.0), False),
        (_make_block(gas_used=True), False),
        (_make_block(timestamp=-1), False),
        (_make_block(timestamp=1.0), False),
        (_make_block(timestamp=True), False),
        (_make_block(base_fee_per_gas=1000000000), True),
        (_make_block(base_fee_per_gas=-1000000000), False),
        (_make_block(base_fee_per_gas=1000000000.0), False),
        (_make_block(base_fee_per_gas="1000000000"), False),
        (_make_block(uncles=[ZERO_32BYTES]), True),
        (_make_block(uncles=[ZERO_32BYTES, HASH32_AS_TEXT]), False),
        (_make_block(transactions=[ZERO_32BYTES]), True),
        (_make_block(transactions=[_make_legacy_txn()]), True),
        (_make_block(transactions=[ZERO_32BYTES, _make_legacy_txn()]), False),
        (_make_block(transactions=[_make_access_list_txn()]), True),
        (_make_block(transactions=[ZERO_32BYTES, _make_access_list_txn()]), False),
        (_make_block(transactions=[_make_dynamic_fee_txn()]), True),
        (_make_block(transactions=[ZERO_32BYTES, _make_dynamic_fee_txn()]), False),
        (_make_block(transactions=[ZERO_32BYTES, HASH32_AS_TEXT]), False),
        (_make_block(withdrawals=[_make_withdrawal()]), True),
        (_make_block(withdrawals=[_make_withdrawal(address=ADDRESS_A)]), True),
        (_make_block(withdrawals=[_make_withdrawal(index=-1)]), False),
        (_make_block(withdrawals=[_make_withdrawal(index=2**64)]), False),
        (_make_block(withdrawals=[_make_withdrawal(validator_index=-1)]), False),
        (_make_block(withdrawals=[_make_withdrawal(validator_index=2**64)]), False),
        (_make_block(withdrawals=[_make_withdrawal(amount=-1)]), False),
        (_make_block(withdrawals=[_make_withdrawal(amount=2**64)]), False),
        (_make_block(withdrawals=[_make_withdrawal(address="1")]), False),
        (
            _make_block(
                withdrawals=[_make_withdrawal(address=encode_hex(ZERO_ADDRESS))]
            ),
            False,
        ),
    ),
)
def test_block_output_validation(validator, block, is_valid):
    if is_valid:
        validator.validate_outbound_block(block)
    else:
        with pytest.raises(ValidationError):
            validator.validate_outbound_block(block)


@pytest.mark.parametrize(
    "accounts,is_valid",
    (
        ([ADDRESS_A], True),
        ([ADDRESS_A, encode_hex(ADDRESS_A)], False),
    ),
)
def test_accounts_output_validation(validator, accounts, is_valid):
    if is_valid:
        validator.validate_outbound_accounts(accounts)
    else:
        with pytest.raises(ValidationError):
            validator.validate_outbound_accounts(accounts)



from __future__ import (
    unicode_literals,
)

import binascii
from typing import (
    Dict,
    List,
    Union,
)

from eth_typing import (
    HexStr,
)
from eth_utils import (
    decode_hex,
    is_boolean,
    is_bytes,
    is_checksum_address,
    is_checksum_formatted_address,
    is_dict,
    is_hex,
    is_hex_address,
    is_hexstr,
    is_integer,
    is_list_like,
    is_string,
    is_text,
    remove_0x_prefix,
)
from eth_utils.toolz import (
    partial,
)

from eth_tester.constants import (
    BLOB_TRANSACTION_PARAMS,
    BLOCK_NUMBER_META_VALUES,
)
from eth_tester.exceptions import (
    ValidationError,
)

from .common import (
    validate_address,
    validate_dict,
    validate_positive_integer,
    validate_text,
    validate_transaction_type,
    validate_uint8,
    validate_uint64,
    validate_uint256,
)


def is_32_bytes(value):
    return is_bytes(value) and len(value) == 32


def is_32byte_hex_string(value):
    return is_text(value) and is_hex(value) and len(remove_0x_prefix(value)) == 64


def is_topic(value):
    return value is None or is_32byte_hex_string(value) or is_32_bytes(value)


def validate_32_byte_hex_value(value, name):
    error_message = (
        "{} must be a hexadecimal encoded 32 byte string.  Got: "
        "{}".format(name, value)
    )
    if not is_32byte_hex_string(value):
        raise ValidationError(error_message)


MAX_TIMESTAMP = 33040162800  # January 1st 3017 is appropriately far in the future.


def validate_timestamp(value):
    validate_positive_integer(value)

    if value >= MAX_TIMESTAMP:
        raise ValidationError(
            "Timestamp values must be less than {}.  Got {}".format(
                MAX_TIMESTAMP,
                value,
            )
        )


def validate_block_number(value):
    error_message = (
        "Block number must be a positive integer or one of the strings "
        "'latest', 'earliest', or 'pending'.  Got: {}".format(value)
    )
    if is_string(value):
        validate_text(value)
        if value not in BLOCK_NUMBER_META_VALUES:
            raise ValidationError(error_message)
    elif not is_integer(value) or is_boolean(value):
        raise ValidationError(error_message)
    elif value < 0:
        raise ValidationError(error_message)


validate_block_hash = partial(validate_32_byte_hex_value, name="Block hash")
validate_transaction_hash = partial(validate_32_byte_hex_value, name="Transaction hash")
validate_filter_id = partial(validate_positive_integer)


def validate_account(value):
    if not is_text(value) or not is_hex_address(value):
        raise ValidationError(
            f"Address must be 20 bytes encoded as hexadecimal - address: {value}"
        )
    elif is_checksum_formatted_address(value) and not is_checksum_address(value):
        raise ValidationError("Address does not validate EIP55 checksum")


def validate_inbound_storage_slot(value):
    error_msg = (
        "Storage slot must be a hex string representation of a positive integer - "
        f"slot: {value}"
    )
    if not (is_hexstr(value) and value.startswith("0x")):
        raise ValidationError(error_msg)

    try:
        int_val = int(value, 16)
    except ValueError:
        raise ValidationError(error_msg)

    validate_uint256(int_val)


def is_valid_topic_array(value):
    if not is_list_like(value):
        return False
    return all(
        is_valid_topic_array(item) if is_list_like(item) else is_topic(item)
        for item in value
    )


def validate_filter_params(from_block, to_block, address, topics):
    # blocks
    if from_block is not None:
        validate_block_number(from_block)
    if to_block is not None:
        validate_block_number(to_block)

    # address
    if address is None:
        pass
    elif is_list_like(address):
        if not address:
            raise ValidationError(
                "Address must be either a single hexadecimal encoded address or "
                "a non-empty list of hexadecimal encoded addresses"
            )
        for sub_address in address:
            validate_account(sub_address)
    elif not is_hex_address(address):
        validate_account(address)

    invalid_topics_message = (
        "Topics must be one of `None` or an array of topics. Each topic must be 32 "
        "bytes, represented as a bytestring or its hex string equivalent. A "
        'filter query of topics using "OR" can be achieved using a sub-array of '
        "topics. See https://eth.wiki/json-rpc/API#eth_newfilter for more details."
    )
    # topics
    if topics is None:
        pass
    elif not is_list_like(topics):
        raise ValidationError(invalid_topics_message)
    elif is_valid_topic_array(topics):
        return True
    else:
        raise ValidationError(invalid_topics_message)


def validate_private_key(value):
    if (
        not is_text(value)
        or not is_hex(value)
        or not len(remove_0x_prefix(value)) == 64
    ):
        raise ValidationError("Private keys must be 32 bytes encoded as hexadecimal")


TRANSACTION_KEYS = {
    "type",
    "chain_id",
    "from",
    "to",
    "gas",
    "gas_price",
    "max_fee_per_gas",
    "max_priority_fee_per_gas",
    "value",
    "data",
    "access_list",
    "nonce",
}

SIGNED_TRANSACTION_KEYS = {
    "r",
    "s",
    "v",
}

TRANSACTION_INTERNAL_TYPE_INFO = {
    "send": TRANSACTION_KEYS,
    "send_signed": TRANSACTION_KEYS.union(SIGNED_TRANSACTION_KEYS),
    "call": TRANSACTION_KEYS,
    "estimate": TRANSACTION_KEYS,
}

ALLOWED_TRANSACTION_INTERNAL_TYPES = set(TRANSACTION_INTERNAL_TYPE_INFO.keys())


def validate_transaction(value, txn_internal_type):
    if txn_internal_type not in ALLOWED_TRANSACTION_INTERNAL_TYPES:
        raise TypeError(
            "the `txn_internal_type` parameter must be one of send/call/estimate"
        )
    if not is_dict(value):
        raise ValidationError(
            "Transaction must be a dictionary. Got: {}".format(type(value))
        )

    unknown_keys = tuple(
        sorted(
            set(value.keys()).difference(
                TRANSACTION_INTERNAL_TYPE_INFO[txn_internal_type],
            )
        )
    )
    if unknown_keys:
        if any(k in value for k in BLOB_TRANSACTION_PARAMS):
            raise ValidationError(
                "Transaction contains blob-specific parameters. Blob transactions are "
                "only supported via `eth_sendRawTransaction`, rlp encoding the blob "
                "sidecar data along with the transaction as per the EIP-4844 "
                "`PooledTransaction` model."
            )
        raise ValidationError(
            "Only the keys '{}' are allowed.  Got extra keys: '{}'".format(
                "/".join(
                    tuple(sorted(TRANSACTION_INTERNAL_TYPE_INFO[txn_internal_type]))
                ),
                "/".join(unknown_keys),
            )
        )

    if txn_internal_type == "send":
        required_keys = {"from", "gas"}
    elif txn_internal_type == "send_signed":
        required_keys = {"from", "gas"} | SIGNED_TRANSACTION_KEYS
    elif txn_internal_type in {"estimate", "call"}:
        required_keys = {"from"}
    else:
        raise Exception("Invariant: code path should be unreachable")

    missing_required_keys = tuple(sorted(required_keys.difference(value.keys())))
    if missing_required_keys:
        raise ValidationError(
            "Transaction is missing the required keys: '{}'".format(
                "/".join(missing_required_keys),
            )
        )

    if "type" in value:
        # type is validated but not required. If this value exists, it will be popped
        # out of the dict and the type will instead be inferred from the
        # transaction params.
        validate_transaction_type(value["type"])

    if "from" in value:
        validate_account(value["from"])

    if "to" in value and value["to"] != "":
        validate_account(value["to"])
    elif "to" in value and value["to"] == "":
        validate_text(value["to"])

    if "gas" in value:
        validate_uint256(value["gas"])

    if "gas_price" in value:
        validate_uint256(value["gas_price"])

    if "max_fee_per_gas" in value:
        validate_uint256(value["max_fee_per_gas"])
        if "gas_price" in value:
            raise ValidationError("Mixed legacy and dynamic fee transaction values")

    if "max_priority_fee_per_gas" in value:
        validate_uint256(value["max_priority_fee_per_gas"])
        if "gas_price" in value:
            raise ValidationError("Mixed legacy and dynamic fee transaction values")

    if "value" in value:
        validate_uint256(value["value"])

    if "nonce" in value:
        validate_uint256(value["nonce"])

    if "data" in value:
        bad_data_message = (
            "Transaction 'data' must be a hexadecimal encoded string.  Got: "
            "{}".format(value["data"])
        )
        if not is_text(value["data"]):
            raise ValidationError(bad_data_message)
        elif not remove_0x_prefix(value["data"]):
            pass
        elif not is_hex(value["data"]):
            raise ValidationError(bad_data_message)
        try:
            decode_hex(value["data"])
        except (binascii.Error, TypeError):
            # TypeError is for python2
            # binascii.Error is for python3
            raise ValidationError(bad_data_message)

    if "access_list" in value:
        _validate_inbound_access_list(value["access_list"])

    if txn_internal_type == "send_signed":
        validate_uint256(value["r"])
        validate_uint256(value["s"])
        validate_uint8(value["v"])


def _validate_inbound_access_list(access_list):
    """
    Validates the structure of an inbound access list. This is similar to the JSON-RPC
    structure for an access list only with `under_score` keys rather than `camelCase`.

    >>> _access_list = (
    ...     {
    ...         'address': '0xde0b295669a9fd93d5f28d9ec85e40f4cb697bae',
    ...         'storage_keys': (
    ...             '0x0000000000000000000000000000000000000000000000000000000000000003',  # noqa: E501
    ...             '0x0000000000000000000000000000000000000000000000000000000000000007',  # noqa: E501
    ...         )
    ...     },
    ...     {
    ...         'address': '0xbb9bc244d798123fde783fcc1c72d3bb8c189413',
    ...         'storage_keys': ()
    ...     },
    ... )
    """
    if not is_list_like(access_list):
        raise ValidationError("access_list is not list-like")
    for entry in access_list:
        if not is_dict(entry) and len(entry) != 2:
            raise ValidationError(f"access_list entry not properly formatted: {entry}")
        address = entry.get("address")
        storage_keys = entry.get("storage_keys")
        if not is_hex_address(address):
            raise ValidationError(
                f"access_list address must be a hexadecimal address: {address}"
            )
        if not is_list_like(storage_keys):
            raise ValidationError(
                f"access_list storage keys are not list-like: {storage_keys}"
            )
        if len(storage_keys) > 0 and not all(
            is_32byte_hex_string(k) for k in storage_keys
        ):
            raise ValidationError(
                "one or more access list storage keys not formatted "
                f"properly: {storage_keys}"
            )


def validate_raw_transaction(raw_transaction):
    if not is_text(raw_transaction) or not is_hex(raw_transaction):
        raise ValidationError(
            "Raw Transaction must be a hexadecimal encoded string.  Got: "
            "{}".format(raw_transaction)
        )


INBOUND_WITHDRAWAL_VALIDATORS = {
    "index": validate_uint64,
    "validator_index": validate_uint64,
    "address": validate_address,
    "amount": validate_uint64,
}


def validate_inbound_withdrawals(
    withdrawals_list: List[Dict[str, Union[int, str, HexStr, bytes]]],
):
    if len(withdrawals_list) == 0:
        raise ValidationError("Withdrawals list must not be empty.")

    for withdrawal_dict in withdrawals_list:
        validate_dict(withdrawal_dict, key_validators=INBOUND_WITHDRAWAL_VALIDATORS)



from __future__ import (
    unicode_literals,
)

from eth_utils import (
    big_endian_to_int,
    is_list_like,
)
import pytest

from eth_tester.utils.filters import (
    check_if_address_match,
    check_if_from_block_match,
    check_if_log_matches,
    check_if_to_block_match,
    check_if_topics_match,
    check_single_topic_match,
    extrapolate_flat_topic_from_topic_list,
    is_flat_topic_array,
    is_topic,
    is_topic_array,
    is_valid_with_nested_topic_array,
)

TOPIC_A = b"\x00" * 32
TOPIC_B = b"\x00" * 31 + b"\x01"
TOPIC_C = b"\x00" * 31 + b"\x02"


def topic_to_name(param):
    if is_list_like(param):
        return [topic_to_name(t) for t in param]
    if isinstance(param, bytes):
        topic_int = big_endian_to_int(param)
        return chr(topic_int + 65)
    else:
        return param


def topic_id(param):
    return repr(topic_to_name(param))


@pytest.mark.parametrize(
    "value,expected",
    (
        # bad values
        ("", False),
        ("a", False),
        (1, False),
        (True, False),
        ({"a": 1, "b": 2}, False),
        (tuple(), False),
        (list(), False),
        (("a", "b"), False),
        (["a", "b"], False),
        (b"", False),
        (b"arst", False),
        # good values
        (None, True),
        (TOPIC_A, True),
        (TOPIC_B, True),
    ),
)
def test_is_topic(value, expected):
    actual = is_topic(value)
    assert actual is expected


TOPICS_EMPTY = tuple()
TOPICS_SINGLE_NULL = (None,)
TOPICS_MANY = (TOPIC_A, TOPIC_B)
TOPICS_MANY_WITH_NULL = (TOPIC_A, None, TOPIC_B)


@pytest.mark.parametrize(
    "value,expected",
    (
        # bad values
        ("", False),
        ("a", False),
        (1, False),
        (True, False),
        ({"a": 1, "b": 2}, False),
        (None, False),
        (b"", False),
        (b"arst", False),
        (("a", "b"), False),
        (["a", "b"], False),
        ([b"a", b"b"], False),
        ([b"a", None, b"b"], False),
        (list(), False),
        ([None], False),
        ((None, b"a"), False),
        ((TOPIC_A, b"a"), False),
        ((b"a", None), False),
        ((b"a", TOPIC_A), False),
        ((TOPIC_A, b"a", TOPIC_B), False),
        # good values
        (TOPICS_EMPTY, True),
        (TOPICS_SINGLE_NULL, True),
        (TOPICS_MANY, True),
        (TOPICS_MANY_WITH_NULL, True),
    ),
)
def test_is_valid_topic_array_with_flat_topic_arrays(value, expected):
    actual = is_flat_topic_array(value)
    assert actual is expected


NESTED_TOPICS_A = (TOPICS_EMPTY,)
NESTED_TOPICS_B = (TOPICS_EMPTY, TOPICS_SINGLE_NULL)
NESTED_TOPICS_C = (TOPICS_SINGLE_NULL, TOPICS_MANY)
NESTED_TOPICS_D = (TOPICS_MANY_WITH_NULL, TOPICS_MANY, TOPICS_EMPTY)
NESTED_TOPICS_E = (TOPIC_A, TOPICS_MANY, TOPICS_EMPTY)


@pytest.mark.parametrize(
    "value,expected",
    (
        # bad values
        ("", False),
        ("a", False),
        (1, False),
        (True, False),
        ({"a": 1, "b": 2}, False),
        (None, False),
        (b"", False),
        (b"arst", False),
        (("a", "b"), False),
        (["a", "b"], False),
        ([b"a", b"b"], False),
        ([b"a", None, b"b"], False),
        (list(), False),
        ([None], False),
        (([],), False),
        (([tuple()],), False),
        ([tuple()], False),
        ((tuple(), []), False),
        ((TOPICS_EMPTY, (b"arst",)), False),
        (TOPIC_A, False),
        (TOPICS_EMPTY, False),
        # good values
        (TOPICS_SINGLE_NULL, True),
        (TOPICS_MANY, True),
        (TOPICS_MANY_WITH_NULL, True),
        (NESTED_TOPICS_A, True),
        (NESTED_TOPICS_B, True),
        (NESTED_TOPICS_C, True),
        (NESTED_TOPICS_D, True),
        (NESTED_TOPICS_E, True),
    ),
)
def test_is_valid_with_nested_topic_array(value, expected):
    actual = is_valid_with_nested_topic_array(value)
    assert actual is expected


@pytest.mark.parametrize(
    "value,expected",
    (
        # bad values
        ("", False),
        ("a", False),
        (1, False),
        (True, False),
        ({"a": 1, "b": 2}, False),
        (None, False),
        (b"", False),
        (b"arst", False),
        (("a", "b"), False),
        (["a", "b"], False),
        ([b"a", b"b"], False),
        ([b"a", None, b"b"], False),
        (list(), False),
        ([None], False),
        (([],), False),
        (([tuple()],), False),
        ([tuple()], False),
        ((tuple(), []), False),
        ((TOPICS_EMPTY, (b"arst",)), False),
        # good values
        (TOPICS_EMPTY, True),
        (TOPICS_SINGLE_NULL, True),
        (TOPICS_MANY, True),
        (TOPICS_MANY_WITH_NULL, True),
        (NESTED_TOPICS_A, True),
        (NESTED_TOPICS_B, True),
        (NESTED_TOPICS_C, True),
        (NESTED_TOPICS_D, True),
    ),
)
def test_is_topic_array(value, expected):
    actual = is_topic_array(value)
    assert actual is expected


TOPIC_A_AS_TEXT = "\x00" * 32
TOPIC_B_AS_TEXT = "\x00" * 31 + "\x01"


@pytest.mark.parametrize(
    "value,topic,expected",
    (
        # bad values
        ("mismatch", TOPIC_A, False),
        (TOPIC_A_AS_TEXT, TOPIC_A, False),
        (TOPIC_B, TOPIC_A, False),
        # good values
        (TOPIC_A, TOPIC_A, True),
        (TOPIC_B, TOPIC_B, True),
        (TOPIC_A, None, True),
        (TOPIC_B, None, True),
    ),
)
def test_check_single_topic_match(value, topic, expected):
    actual = check_single_topic_match(value, topic)
    assert actual is expected


@pytest.mark.parametrize(
    "block_number,_type,from_block,expected",
    (
        # bad values
        (10, "mined", 11, False),
        (10, "mined", "pending", False),
        (10, "mined", "earliest", False),
        # good values
        (10, "mined", None, True),
        (10, "mined", 10, True),
        (20, "mined", 10, True),
        (10, "mined", "latest", True),
        (10, "pending", "pending", True),
        (10, "pending", "earliest", True),
    ),
)
def test_check_if_from_block_match(block_number, _type, from_block, expected):
    actual = check_if_from_block_match(block_number, _type, from_block)
    assert actual is expected


@pytest.mark.parametrize(
    "block_number,_type,to_block,expected",
    (
        # bad values
        (11, "mined", 10, False),
        (10, "mined", "pending", False),
        (10, "mined", "earliest", False),
        # good values
        (10, "mined", None, True),
        (10, "mined", 10, True),
        (9, "mined", 10, True),
        (10, "mined", "latest", True),
        (10, "pending", "pending", True),
        (10, "pending", "earliest", True),
    ),
)
def test_check_if_to_block_match(block_number, _type, to_block, expected):
    actual = check_if_to_block_match(block_number, _type, to_block)
    assert actual is expected


TOPICS_EMPTY = tuple()
TOPICS_ONLY_A = (TOPIC_A,)
TOPICS_ONLY_B = (TOPIC_B,)
TOPICS_ONLY_C = (TOPIC_C,)
TOPICS_A_A = (TOPIC_A, TOPIC_A)
TOPICS_A_B = (TOPIC_A, TOPIC_B)
TOPICS_A_C = (TOPIC_A, TOPIC_C)
TOPICS_A_B_C = (TOPIC_A, TOPIC_B, TOPIC_C)
TOPICS_A_C_B = (TOPIC_A, TOPIC_C, TOPIC_B)
TOPICS_B_A = (TOPIC_B, TOPIC_A)
TOPICS_B_C = (TOPIC_B, TOPIC_C)
TOPICS_B_A_C = (TOPIC_B, TOPIC_A, TOPIC_C)
TOPICS_B_C_A = (TOPIC_B, TOPIC_C, TOPIC_A)


FILTER_MATCH_ALL = tuple()
FILTER_MATCH_ONE_OR_MORE = (None,)
FILTER_MATCH_TWO_OR_MORE = (None, None)
FILTER_MATCH_THREE_OR_MORE = (None, None, None)
FILTER_MATCH_A = (TOPIC_A,)
FILTER_MATCH_B = (TOPIC_B,)
FILTER_MATCH_C = (TOPIC_C,)
FILTER_MATCH_A_ANY = (TOPIC_A, None)
FILTER_MATCH_B_ANY = (TOPIC_B, None)
FILTER_MATCH_C_ANY = (TOPIC_C, None)
FILTER_MATCH_ANY_A = (None, TOPIC_A)
FILTER_MATCH_ANY_B = (None, TOPIC_B)
FILTER_MATCH_ANY_C = (None, TOPIC_C)
FILTER_MATCH_A_B = (TOPIC_A, TOPIC_B)
FILTER_MATCH_B_C = (TOPIC_B, TOPIC_C)
FILTER_MATCH_A_B_C = (TOPIC_A, TOPIC_B, TOPIC_C)
FILTER_MATCH_A_C_B = (TOPIC_A, TOPIC_C, TOPIC_B)


@pytest.mark.parametrize(
    "log_topics,filter_topics,expected",
    (
        # match all values
        (TOPICS_EMPTY, FILTER_MATCH_ALL, True),
        (TOPICS_ONLY_A, FILTER_MATCH_ALL, True),
        (TOPICS_ONLY_B, FILTER_MATCH_ALL, True),
        (TOPICS_ONLY_C, FILTER_MATCH_ALL, True),
        (TOPICS_A_A, FILTER_MATCH_ALL, True),
        (TOPICS_A_B, FILTER_MATCH_ALL, True),
        (TOPICS_A_C, FILTER_MATCH_ALL, True),
        (TOPICS_B_C, FILTER_MATCH_ALL, True),
        (TOPICS_B_A, FILTER_MATCH_ALL, True),
        (TOPICS_A_B_C, FILTER_MATCH_ALL, True),
        (TOPICS_A_C_B, FILTER_MATCH_ALL, True),
        (TOPICS_B_A_C, FILTER_MATCH_ALL, True),
        (TOPICS_B_C_A, FILTER_MATCH_ALL, True),
        # length 1 matches
        (TOPICS_EMPTY, FILTER_MATCH_ONE_OR_MORE, False),
        (TOPICS_ONLY_A, FILTER_MATCH_ONE_OR_MORE, True),
        (TOPICS_ONLY_B, FILTER_MATCH_ONE_OR_MORE, True),
        (TOPICS_ONLY_C, FILTER_MATCH_ONE_OR_MORE, True),
        (TOPICS_EMPTY, FILTER_MATCH_A, False),
        (TOPICS_EMPTY, FILTER_MATCH_B, False),
        (TOPICS_EMPTY, FILTER_MATCH_C, False),
        (TOPICS_ONLY_A, FILTER_MATCH_A, True),
        (TOPICS_ONLY_B, FILTER_MATCH_B, True),
        (TOPICS_ONLY_C, FILTER_MATCH_C, True),
        (TOPICS_ONLY_B, FILTER_MATCH_A, False),
        (TOPICS_ONLY_C, FILTER_MATCH_A, False),
        (TOPICS_ONLY_A, FILTER_MATCH_B, False),
        (TOPICS_ONLY_C, FILTER_MATCH_B, False),
        (TOPICS_ONLY_A, FILTER_MATCH_C, False),
        (TOPICS_ONLY_B, FILTER_MATCH_C, False),
        (TOPICS_A_A, FILTER_MATCH_A, True),
        (TOPICS_A_B, FILTER_MATCH_A, True),
        (TOPICS_A_C, FILTER_MATCH_A, True),
        (TOPICS_A_B_C, FILTER_MATCH_A, True),
        (TOPICS_A_C_B, FILTER_MATCH_A, True),
        (TOPICS_B_A, FILTER_MATCH_A, False),
        (TOPICS_B_C, FILTER_MATCH_A, False),
        (TOPICS_B_A_C, FILTER_MATCH_A, False),
        (TOPICS_B_C_A, FILTER_MATCH_A, False),
        # length 2 matches
        (TOPICS_EMPTY, FILTER_MATCH_TWO_OR_MORE, False),
        (TOPICS_A_A, FILTER_MATCH_TWO_OR_MORE, True),
        (TOPICS_A_B, FILTER_MATCH_TWO_OR_MORE, True),
        (TOPICS_ONLY_A, FILTER_MATCH_TWO_OR_MORE, False),
        (TOPICS_ONLY_B, FILTER_MATCH_TWO_OR_MORE, False),
        (TOPICS_ONLY_C, FILTER_MATCH_TWO_OR_MORE, False),
        (TOPICS_A_A, FILTER_MATCH_A_B, False),
        (TOPICS_A_B, FILTER_MATCH_A_B, True),
        (TOPICS_A_B_C, FILTER_MATCH_A_B, True),
        (TOPICS_A_C, FILTER_MATCH_A_B, False),
        (TOPICS_A_C, FILTER_MATCH_B_C, False),
        (TOPICS_B_C, FILTER_MATCH_B_C, True),
        (TOPICS_B_C_A, FILTER_MATCH_B_C, True),
        (TOPICS_A_A, FILTER_MATCH_A_ANY, True),
        (TOPICS_A_B, FILTER_MATCH_A_ANY, True),
        (TOPICS_A_C, FILTER_MATCH_A_ANY, True),
        (TOPICS_A_B_C, FILTER_MATCH_A_ANY, True),
        (TOPICS_A_C_B, FILTER_MATCH_A_ANY, True),
        (TOPICS_B_C, FILTER_MATCH_A_ANY, False),
        (TOPICS_A_B, FILTER_MATCH_B_ANY, False),
        (TOPICS_A_C, FILTER_MATCH_B_ANY, False),
        (TOPICS_B_C, FILTER_MATCH_B_ANY, True),
        (TOPICS_B_C_A, FILTER_MATCH_B_ANY, True),
        (TOPICS_B_A, FILTER_MATCH_B_ANY, True),
        (TOPICS_B_A_C, FILTER_MATCH_B_ANY, True),
        (TOPICS_A_A, FILTER_MATCH_ANY_A, True),
        (TOPICS_A_B, FILTER_MATCH_ANY_A, False),
        (TOPICS_B_A, FILTER_MATCH_ANY_A, True),
        (TOPICS_B_A_C, FILTER_MATCH_ANY_A, True),
        (TOPICS_A_B, FILTER_MATCH_ANY_B, True),
        (TOPICS_A_B_C, FILTER_MATCH_ANY_B, True),
        (TOPICS_A_B, FILTER_MATCH_ANY_C, False),
        (TOPICS_A_B_C, FILTER_MATCH_ANY_C, False),
        # length 3 matches
        (TOPICS_EMPTY, FILTER_MATCH_THREE_OR_MORE, False),
        (TOPICS_A_B_C, FILTER_MATCH_THREE_OR_MORE, True),
        (TOPICS_A_C_B, FILTER_MATCH_THREE_OR_MORE, True),
        (TOPICS_B_A_C, FILTER_MATCH_THREE_OR_MORE, True),
        (TOPICS_B_C_A, FILTER_MATCH_THREE_OR_MORE, True),
        (TOPICS_A_A, FILTER_MATCH_THREE_OR_MORE, False),
        (TOPICS_A_B_C, FILTER_MATCH_A_B_C, True),
        (TOPICS_A_C_B, FILTER_MATCH_A_B_C, False),
        (TOPICS_A_C_B, FILTER_MATCH_A_C_B, True),
        (TOPICS_A_B_C, FILTER_MATCH_A_C_B, False),
        # positional topic options matches
        (TOPICS_EMPTY, (FILTER_MATCH_A, FILTER_MATCH_B, FILTER_MATCH_C), False),
        (TOPICS_ONLY_A, (FILTER_MATCH_A, FILTER_MATCH_B, FILTER_MATCH_C), False),
        (TOPICS_ONLY_B, (FILTER_MATCH_A, FILTER_MATCH_B, FILTER_MATCH_C), False),
        (TOPICS_ONLY_C, (FILTER_MATCH_A, FILTER_MATCH_B, FILTER_MATCH_C), False),
        (TOPICS_A_B, (FILTER_MATCH_B, FILTER_MATCH_C), False),
        (TOPICS_A_B, (FILTER_MATCH_B, FILTER_MATCH_C, FILTER_MATCH_A), False),
        (TOPICS_A_C, (FILTER_MATCH_A_ANY, FILTER_MATCH_ANY_A), True),
        (TOPICS_B_A, (FILTER_MATCH_A_ANY, FILTER_MATCH_ANY_A), True),
        (TOPICS_B_C, (FILTER_MATCH_A_ANY, FILTER_MATCH_ANY_A), True),
        (TOPICS_A_B_C, (FILTER_MATCH_A, FILTER_MATCH_A_B_C, FILTER_MATCH_C), True),
        (TOPICS_A_B_C, (FILTER_MATCH_A, FILTER_MATCH_A_C_B, FILTER_MATCH_C), True),
        (TOPICS_A_B_C, (FILTER_MATCH_A, FILTER_MATCH_A_C_B, TOPIC_C), True),
    ),
    ids=topic_id,
)
def test_check_if_topics_match(log_topics, filter_topics, expected):
    actual = check_if_topics_match(log_topics, filter_topics)
    assert actual is expected


ADDRESS_A = b"\x00" * 20
ADDRESS_B = b"\x00" * 19 + b"\x01"
ADDRESS_C = b"\x00" * 19 + b"\x02"
ADDRESS_D = b"\x00" * 19 + b"\x03"


@pytest.mark.parametrize(
    "address,addresses,expected",
    (
        (ADDRESS_A, None, True),
        (ADDRESS_B, None, True),
        (ADDRESS_C, None, True),
        (ADDRESS_D, None, True),
        (ADDRESS_A, ADDRESS_A, True),
        (ADDRESS_B, ADDRESS_A, False),
        (ADDRESS_C, ADDRESS_A, False),
        (ADDRESS_D, ADDRESS_A, False),
        (ADDRESS_A, (ADDRESS_A,), True),
        (ADDRESS_B, (ADDRESS_A,), False),
        (ADDRESS_C, (ADDRESS_A,), False),
        (ADDRESS_D, (ADDRESS_A,), False),
        (ADDRESS_A, (ADDRESS_A, ADDRESS_B), True),
        (ADDRESS_B, (ADDRESS_A, ADDRESS_B), True),
        (ADDRESS_C, (ADDRESS_A, ADDRESS_B), False),
        (ADDRESS_D, (ADDRESS_A, ADDRESS_B), False),
        (ADDRESS_A, (ADDRESS_B, ADDRESS_A), True),
        (ADDRESS_B, (ADDRESS_B, ADDRESS_A), True),
        (ADDRESS_C, (ADDRESS_B, ADDRESS_A), False),
        (ADDRESS_D, (ADDRESS_B, ADDRESS_A), False),
    ),
)
def test_check_if_address_match(address, addresses, expected):
    actual = check_if_address_match(address, addresses)
    assert actual is expected


def _make_log(block_number=10, topics=None, address=ADDRESS_A, _type="mined", **kwargs):
    return dict(
        block_number=block_number,
        topics=topics or tuple(),
        address=address,
        type=_type,
        **kwargs
    )


def _make_filter(from_block=None, to_block=None, topics=None, addresses=None):
    return {
        "from_block": from_block,
        "to_block": to_block,
        "topics": topics,
        "addresses": addresses,
    }


@pytest.mark.parametrize(
    "log_entry,filter_params,expected",
    (
        # block numbers
        (_make_log(), _make_filter(), True),
        (_make_log(block_number=10), _make_filter(from_block=11), False),
        (_make_log(block_number=30), _make_filter(from_block=10), True),
        (_make_log(block_number=30), _make_filter(from_block=10, to_block=20), False),
        (_make_log(block_number=30), _make_filter(to_block=20), False),
        (_make_log(block_number=20), _make_filter(to_block=20), True),
        # topics
        (_make_log(topics=(TOPIC_A,)), _make_filter(topics=FILTER_MATCH_ALL), True),
        (
            _make_log(topics=(TOPIC_A, TOPIC_B)),
            _make_filter(topics=FILTER_MATCH_ALL),
            True,
        ),
        (_make_log(topics=(TOPIC_A,)), _make_filter(topics=FILTER_MATCH_A), True),
        (_make_log(topics=(TOPIC_B,)), _make_filter(topics=FILTER_MATCH_A), False),
        (
            _make_log(topics=(TOPIC_A, TOPIC_A)),
            _make_filter(topics=FILTER_MATCH_A_ANY),
            True,
        ),
        (
            _make_log(topics=(TOPIC_A, TOPIC_B)),
            _make_filter(topics=FILTER_MATCH_A_ANY),
            True,
        ),
        (
            _make_log(topics=(TOPIC_B, TOPIC_A)),
            _make_filter(topics=FILTER_MATCH_A_ANY),
            False,
        ),
        (_make_log(topics=TOPICS_A_B), _make_filter(topics=FILTER_MATCH_A_ANY), True),
        (_make_log(topics=TOPICS_B_A), _make_filter(topics=FILTER_MATCH_A_ANY), False),
        (
            _make_log(topics=TOPICS_A_B),
            _make_filter(topics=(FILTER_MATCH_A_B, FILTER_MATCH_B_C)),
            True,
        ),
        (
            _make_log(topics=TOPICS_B_A),
            _make_filter(topics=(FILTER_MATCH_A_B, FILTER_MATCH_B_C)),
            False,
        ),
        (
            _make_log(topics=TOPICS_B_C),
            _make_filter(topics=(FILTER_MATCH_A_B, FILTER_MATCH_B_C)),
            True,
        ),
    ),
)
def test_check_if_log_matches(log_entry, filter_params, expected):
    actual = check_if_log_matches(log_entry, **filter_params)
    assert actual == expected


@pytest.mark.parametrize(
    "topic_list_input,expected_flat_topics",
    (
        (
            ("A", ("A", "B"), "A"),
            (("A", "A", "A"), ("A", "B", "A")),
        ),
        (
            ("A", ("A", "B", "C"), ("A", "B")),
            (
                ("A", "A", "A"),
                ("A", "A", "B"),
                ("A", "B", "A"),
                ("A", "B", "B"),
                ("A", "C", "A"),
                ("A", "C", "B"),
            ),
        ),
    ),
)
def test_extrapolate_flat_topic_from_topic_list(topic_list_input, expected_flat_topics):
    assert (
        tuple(extrapolate_flat_topic_from_topic_list(topic_list_input))
        == expected_flat_topics
    )



from eth_utils import (
    decode_hex,
)
import pytest

from eth_tester.utils.accounts import (
    private_key_to_address,
)

PK_A = decode_hex("0x58d23b55bc9cdce1f18c2500f40ff4ab7245df9a89505e9b1fa4851f623d241d")
PK_A_ADDRESS = decode_hex("0xdc544d1aa88ff8bbd2f2aec754b1f1e99e1812fd")


@pytest.mark.parametrize(
    "private_key,expected",
    ((PK_A, PK_A_ADDRESS),),
)
def test_private_key_to_address(private_key, expected):
    actual = private_key_to_address(private_key)
    assert actual == expected



import pytest

from eth_tester.normalization import (
    DefaultNormalizer,
)
from tests.utils import (
    make_receipt,
)


@pytest.mark.parametrize(
    "status,contract_address,expected",
    (
        (1, b"\x01" * 20, f"0x{'01' * 20}"),
        (0, b"\x01" * 20, None),
    ),
)
def test_outbound_receipt_contract_address_status_based_normalization(
    status,
    contract_address,
    expected,
):
    receipt = make_receipt(status=status, contract_address=contract_address)

    assert receipt["status"] == status
    assert receipt["contract_address"] == contract_address

    normalized_receipt = DefaultNormalizer.normalize_outbound_receipt(receipt)

    assert normalized_receipt["status"] == status
    assert normalized_receipt["contract_address"] == expected


@pytest.mark.parametrize(
    "value,expected",
    (
        (0, f"0x{'00'*32}"),
        (1, f"0x{'00'*31}01"),
        (2, f"0x{'00'*31}02"),
    ),
)
def test_outbound_storage_normalization(value, expected):
    assert DefaultNormalizer.normalize_outbound_storage(value) == expected



from importlib.metadata import (
    version as __version,
)

from .backends import (
    MockBackend,
    PyEVMBackend,
)
from .main import (
    EthereumTester,
)

__version__ = __version("eth-tester")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import (
    find_packages,
    setup,
)

extras_require = {
    "dev": [
        "build>=0.9.0",
        "bumpversion>=0.5.3",
        "ipython",
        "pre-commit>=3.4.0",
        "tox>=4.0.0",
        "twine",
        "wheel",
    ],
    "docs": [
        "towncrier>=21,<22",
    ],
    "test": [
        "pytest>=7.0.0",
        "pytest-xdist>=2.0.0,<3",
        "eth-hash[pycryptodome]>=0.1.4,<1.0.0",
    ],
    "py-evm": [
        # Pin py-evm to a minor version range to ensure compatibility with the current
        # implemented EVM version.
        "py-evm>=0.10.0b0,<0.11.0b0",
        "eth-hash[pysha3]>=0.1.4,<1.0.0;implementation_name=='cpython'",
        "eth-hash[pycryptodome]>=0.1.4,<1.0.0;implementation_name=='pypy'",
    ],
}

extras_require["dev"] = (
    extras_require["dev"]
    + extras_require["docs"]
    + extras_require["test"]
    + extras_require["py-evm"]
)
# convenience in case someone leaves out the `-`
extras_require["pyevm"] = extras_require["py-evm"]

with open("./README.md") as readme:
    long_description = readme.read()

setup(
    name="eth-tester",
    # *IMPORTANT*: Don't manually change the version here. Use the 'bumpversion' utility.
    version="0.11.0-beta.1",
    description="""eth-tester: Tools for testing Ethereum applications.""",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="The Ethereum Foundation",
    author_email="snakecharmers@ethereum.org",
    url="https://github.com/ethereum/eth-tester",
    include_package_data=True,
    install_requires=[
        "eth-abi>=3.0.1",
        "eth-account>=0.11.2",
        "eth-keys>=0.4.0",
        "eth-utils>=2.0.0",
        "rlp>=3.0.0",
        "semantic_version>=2.6.0",
    ],
    extras_require=extras_require,
    python_requires=">=3.8,<4",
    py_modules=["eth_tester"],
    license="MIT",
    zip_safe=False,
    keywords="ethereum",
    packages=find_packages(exclude=["tests", "tests.*"]),
    package_data={"eth_tester": ["py.typed"]},
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
)



from __future__ import (
    unicode_literals,
)

import pytest

from eth_tester import (
    EthereumTester,
    MockBackend,
)
from eth_tester.utils.backend_testing import (
    BaseTestBackendDirect,
)


@pytest.fixture
def eth_tester():
    backend = MockBackend()
    return EthereumTester(backend=backend)


class TestMockBackendDirect(BaseTestBackendDirect):
    supports_evm_execution = False

    @pytest.mark.skip(reason="receipt status not supported in MockBackend")
    def test_get_transaction_receipt_byzantium(self, eth_tester, test_transaction):
        pass



import pytest

from eth_tester.utils.transactions import (
    remove_matching_transaction_from_list,
)

TX1 = {"from": "0x1", "to": "0x5", "value": 1, "nonce": 0}
TX2 = {"from": "0x2", "to": "0x5", "value": 1, "nonce": 0}
TX3 = {"from": "0x3", "to": "0x5", "value": 1, "nonce": 0}


@pytest.mark.parametrize(
    "tx_list,transaction,expected",
    (
        ([TX1, TX2, TX3], TX2, [TX1, TX3]),
        (
            [TX1, TX2, TX3],
            {"from": "0x2", "to": "0x6", "value": 2, "nonce": 0},
            [TX1, TX3],
        ),
    ),
)
def test_remove_matching_transaction_from_list(tx_list, transaction, expected):
    result = remove_matching_transaction_from_list(tx_list, transaction)
    assert result == expected



from __future__ import (
    absolute_import,
)

from eth_utils import (
    to_bytes,
)
from eth_utils.curried import (
    apply_one_of_formatters,
    decode_hex,
    is_address,
    is_hex,
    is_list_like,
    is_string,
    to_canonical_address,
    to_tuple,
)
from eth_utils.toolz import (
    identity,
    partial,
)

from eth_tester.validation.inbound import (
    is_32_bytes,
    is_32byte_hex_string,
    is_valid_topic_array,
)

from .common import (
    normalize_array,
    normalize_dict,
    normalize_if,
)


def normalize_topic(topic):
    if topic is None:
        return None
    if is_32byte_hex_string(topic):
        return decode_hex(topic)
    if is_32_bytes(topic):
        return topic

    raise TypeError(f"Topic is not in a recognized format: {topic}")


@to_tuple
def normalize_topic_list(topics):
    for topic in topics:
        yield normalize_topic(topic)


@to_tuple
def normalize_filter_params(from_block, to_block, address, topics):
    yield from_block
    yield to_block

    if address is None:
        yield address
    elif is_address(address):
        yield to_canonical_address(address)
    elif is_list_like(address):
        yield tuple(to_canonical_address(item) for item in address)
    else:
        raise TypeError(f"Address is not in a recognized format: {address}")

    if topics is None:
        yield topics
    elif is_valid_topic_array(topics):
        yield tuple(
            normalize_topic_list(item) if is_list_like(item) else normalize_topic(item)
            for item in topics
        )
    else:
        raise TypeError(f"Topics are not in a recognized format: {address}")


def normalize_private_key(value):
    return decode_hex(value)


to_empty_or_canonical_address = apply_one_of_formatters(
    (
        (lambda addr: addr == "", lambda addr: b""),
        (is_hex, to_canonical_address),
    )
)


def _normalize_inbound_access_list(access_list):
    return tuple(
        [
            tuple(
                [
                    to_bytes(hexstr=entry["address"]),
                    tuple([int(k, 16) for k in entry["storage_keys"]]),
                ]
            )
            for entry in access_list
        ]
    )


TRANSACTION_NORMALIZERS = {
    "type": identity,
    "chain_id": identity,
    "from": to_canonical_address,
    "to": to_empty_or_canonical_address,
    "gas": identity,
    "gas_price": identity,
    "max_fee_per_gas": identity,
    "max_priority_fee_per_gas": identity,
    "nonce": identity,
    "value": identity,
    "data": decode_hex,
    "access_list": _normalize_inbound_access_list,
    "r": identity,
    "s": identity,
    "v": identity,
}
normalize_transaction = partial(normalize_dict, normalizers=TRANSACTION_NORMALIZERS)


LOG_ENTRY_NORMALIZERS = {
    "type": identity,
    "log_index": identity,
    "transaction_index": identity,
    "transaction_hash": decode_hex,
    "block_hash": partial(
        normalize_if, conditional_fn=is_string, normalizer=decode_hex
    ),
    "block_number": identity,
    "address": to_canonical_address,
    "data": decode_hex,
    "topics": partial(normalize_array, normalizer=decode_hex),
}
normalize_log_entry = partial(normalize_dict, normalizers=LOG_ENTRY_NORMALIZERS)


def normalize_raw_transaction(raw_transaction_hex):
    return decode_hex(raw_transaction_hex)



from eth_utils import (
    encode_hex,
    is_hex,
    to_dict,
    to_list,
)
from eth_utils.toolz import (
    curry,
)

from eth_tester.utils.encoding import (
    int_to_32byte_big_endian,
)


@curry
@to_dict
def normalize_dict(value, normalizers):
    for key, item in value.items():
        normalizer = normalizers[key]
        yield key, normalizer(item)


@curry
@to_list
def normalize_array(value, normalizer):
    """
    Just `map` but it's nice to have it return a consistent type
    (tuple).
    """
    for item in value:
        yield normalizer(item)


@curry
def normalize_if(value, conditional_fn, normalizer):
    if conditional_fn(value):
        return normalizer(value)
    else:
        return value


def to_integer_if_hex(value):
    if is_hex(value):
        return int(value, 16)
    return value


def int_to_32byte_hex(value):
    return encode_hex(int_to_32byte_big_endian(value))
